from enum import Enum
from typing import Callable, List, Tuple, Type, Union

from rekuest_next.definition.guards import cls_is_union
from rekuest_next.scalars import ValidatorFunction
from rekuest_next.structures.model import (
    is_model,
    retrieve_fullfiled_model,
)
from rekuest_next.structures.serialization.predication import predicate_port
from .utils import get_type_hints, is_annotated
import inflection
from rekuest_next.api.schema import (
    PortInput,
    ChildPortInput,
    DefinitionInput,
    NodeKind,
    PortKind,
    PortScope,
    AssignWidgetInput,
    ReturnWidgetInput,
    PortGroupInput,
    EffectInput,
    ValidatorInput,
)
import inspect
from docstring_parser import parse
from rekuest_next.definition.errors import DefinitionError, NonSufficientDocumentation
import datetime as dt
from rekuest_next.structures.registry import (
    StructureRegistry,
)
from typing import (
    Protocol,
    runtime_checkable,
    Optional,
    List,
    Any,
    Dict,
    get_origin,
    get_args,
)
import types
import typing


def is_nullable(cls):
    is_union = get_origin(cls) is types.UnionType or get_origin(cls) is Union

    if is_union:
        for arg in get_args(cls):
            if arg == type(None):
                return True

    return False


def get_non_nullable_variant(cls):
    non_nullable_args = [arg for arg in get_args(cls) if arg != type(None)]
    if len(non_nullable_args) == 1:
        return non_nullable_args[0]
    # We are dealing with a Union so we still use the same class
    # the logic will be handled in the union path
    # TODO: We might want to handle this better
    return cls


def is_union(cls):
    return (
        get_origin(cls) is types.UnionType
        or get_origin(cls) is Union
        and get_args(cls)[1] != type(None)
    )


def is_list(cls):
    """Check if a class is a list



    if cls.__module__ == "typing":
        if hasattr(cls, "_name"):
            # We are dealing with a Typing Var?
            if cls._name == "List":

    Returns:
        _type_: _description_
    """
    return get_origin(cls) == list


def is_dict(cls):
    """Check if a class is a dict

    if cls.__module__ == "typing":
        if hasattr(cls, "_name"):
            if cls._name == "Dict":

    Returns:
        _type_: _description_
    """
    return get_origin(cls) == dict


def get_dict_value_cls(cls):
    return get_args(cls)[1]


def get_list_value_cls(cls):
    return get_args(cls)[0]


def get_non_null_variants(cls):
    return [arg for arg in get_args(cls) if arg != type(None)]


def is_bool(cls, default=None):
    if inspect.isclass(cls):
        return not issubclass(cls, Enum) and issubclass(cls, bool)
    return False


def is_float(cls):
    if inspect.isclass(cls):
        return not issubclass(cls, Enum) and issubclass(cls, float)
    return False


def is_generator_type(cls):
    if cls.__module__ == "typing":
        if hasattr(cls, "_name"):
            if cls._name == "Generator":
                return True
            if cls._name == "AsyncGenerator":
                return True
    return False


def is_int(cls):
    if inspect.isclass(cls):
        return not issubclass(cls, Enum) and issubclass(cls, int)
    return False


def is_str(cls):
    if inspect.isclass(cls):
        return not issubclass(cls, Enum) and issubclass(cls, str)
    return False


def is_datetime(cls):
    if inspect.isclass(cls):
        return not issubclass(cls, Enum) and (issubclass(cls, dt.datetime))
    return False


def is_structure(cls):
    return True


def convert_child_to_childport(
    cls: Type,
    registry: StructureRegistry,
    nullable: bool = False,
    key: str | None = None,
    description: str | None = None,
) -> Tuple[ChildPortInput, Callable]:
    """Converts a element of a annotation to a child port

    Args:
        cls (Type): The type (class or annotation) of the elemtn
        registry (StructureRegistry): The structure registry to use
        nullable (bool, optional): Is this type optional (recursive parameter).
            Defaults to False.
        is_return (bool, optional): Is this a return type?. Defaults to False.
        annotations (List[AnnotationInput], optional): The annotations for this element.
            Defaults to None.

    Returns:
        Tuple[ChildPortInput, WidgetInput, Callable]: The child port, the widget and the
         converter for the default
    """
    if is_model(cls):
        children = []
        convertermap = {}

        full_filled_model = retrieve_fullfiled_model(cls)

        registry.register_as_model(cls, full_filled_model.identifier)

        for arg in full_filled_model.args:
            child, converter = convert_child_to_childport(
                arg.cls,
                registry,
                nullable=False,
                key=arg.key,
                description=arg.description,
            )
            children.append(child)
            convertermap[arg.key] = converter

        return (
            ChildPortInput(
                kind=PortKind.MODEL,
                children=children,
                scope=PortScope.GLOBAL,
                identifier=full_filled_model.identifier,
                nullable=nullable,
                key=key,
                description=full_filled_model.description,
            ),
            lambda default: default.dict(),
        )

    if is_annotated(cls):
        real_type = cls.__args__[0]

        return convert_child_to_childport(
            real_type, registry, nullable=nullable, key=key
        )

    if is_nullable(cls):
        non_nullable = get_non_nullable_variant(cls)
        return convert_child_to_childport(
            non_nullable, registry, nullable=True, key=key
        )

    if is_union(cls):
        variants = get_non_null_variants(cls)
        children = []
        converters = []
        for index, arg in enumerate(variants):
            child, converter = convert_child_to_childport(
                arg, registry, nullable=False, key="variant_" + str(index)
            )
            converters.append(converter)
            children.append(child)

        return ChildPortInput(
            kind=PortKind.UNION,
            scope=PortScope.GLOBAL,
            key=key,
            children=children,
            nullable=nullable,
            description=description,
        )

    if is_list(cls):
        value_cls = get_list_value_cls(cls)
        child, nested_converter = convert_child_to_childport(
            value_cls, registry, nullable=False, key="..."
        )

        return (
            ChildPortInput(
                kind=PortKind.LIST,
                children=[child],
                scope=PortScope.GLOBAL,
                nullable=nullable,
                key=key,
                description=description,
            ),
            lambda default: (
                [nested_converter(ndefault) for ndefault in default]
                if default
                else None
            ),
        )

    if is_dict(cls):
        value_cls = get_dict_value_cls(cls)
        child, nested_converter = convert_child_to_childport(
            value_cls, registry, nullable=False, key="..."
        )
        return (
            ChildPortInput(
                kind=PortKind.DICT,
                children=[child],
                scope=PortScope.GLOBAL,
                nullable=nullable,
                key=key,
                description=description,
            ),
            lambda default: (
                {key: item in nested_converter(item) for key, item in default.items()}
                if default
                else None
            ),
        )

    if is_bool(cls):
        return (
            ChildPortInput(
                kind=PortKind.BOOL,
                nullable=nullable,
                scope=PortScope.GLOBAL,
                key=key,
                description=description,
            ),
            bool,
        )

    if is_float(cls):
        return (
            ChildPortInput(
                kind=PortKind.FLOAT,
                nullable=nullable,
                scope=PortScope.GLOBAL,
                key=key,
                description=description,
            ),
            float,
        )

    if is_int(cls):
        return (
            ChildPortInput(
                kind=PortKind.INT,
                nullable=nullable,
                scope=PortScope.GLOBAL,
                key=key,
                description=description,
            ),
            int,
        )

    if is_datetime(cls):
        return (
            ChildPortInput(
                kind=PortKind.DATE,
                nullable=nullable,
                scope=PortScope.GLOBAL,
                key=key,
                description=description,
            ),
            lambda x: x.isoformat(),
        )

    if is_str(cls):
        return (
            ChildPortInput(
                kind=PortKind.STRING,
                nullable=nullable,
                scope=PortScope.GLOBAL,
                key=key,
                description=description,
            ),
            str,
        )

    if is_structure(cls):
        return registry.get_child_port_and_default_converter_for_cls(
            cls,
            nullable=nullable,
            key=key,
            description=description,
        )

    raise NotImplementedError(f"Could not convert {cls} to a child port")


def convert_object_to_port(
    cls,
    key,
    registry: StructureRegistry,
    assign_widget=None,
    return_widget=None,
    default=None,
    label=None,
    description=None,
    nullable=False,
    validators: Optional[List[ValidatorInput]] = None,
    effects: Optional[List[EffectInput]] = None,
    groups: Optional[List[str]] = None,
) -> PortInput:
    """
    Convert a class to an Port
    """

    if is_generator_type(cls):
        real_type = cls.__args__[0]

        return convert_object_to_port(
            real_type,
            key,
            registry,
            assign_widget=assign_widget,
            default=default,
            label=label,
            effects=effects,
            nullable=nullable,
            groups=groups,
        )

    if is_model(cls):
        children = []
        converters = []
        set_default = default or {}

        full_filled_model = retrieve_fullfiled_model(cls)

        registry.register_as_model(cls, full_filled_model.identifier)

        for arg in full_filled_model.args:
            child, converter = convert_child_to_childport(
                arg.cls,
                registry,
                nullable=False,
                key=arg.key,
                description=arg.description,
            )
            children.append(child)
            if arg.default:
                set_default[arg.key] = converter(arg.default)

        return PortInput(
            kind=PortKind.MODEL,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            children=children,
            label=label,
            default=set_default,
            nullable=nullable,
            effects=effects,
            description=description or full_filled_model.description,
            groups=groups,
            validators=validators,
            identifier=full_filled_model.identifier,
        )

    if is_annotated(cls):
        real_type = cls.__args__[0]

        return convert_object_to_port(
            real_type,
            key,
            registry,
            assign_widget=assign_widget,
            default=default,
            label=label,
            effects=effects,
            nullable=nullable,
            groups=groups,
        )

    if is_list(cls):
        value_cls = get_list_value_cls(cls)
        child, converter = convert_child_to_childport(
            value_cls, registry, nullable=False, key="..."
        )
        return PortInput(
            kind=PortKind.LIST,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            children=[child],
            label=label,
            default=[converter(item) for item in default] if default else None,
            nullable=nullable,
            effects=effects,
            description=description,
            validators=validators,
            groups=groups,
        )

    if is_nullable(cls):
        return convert_object_to_port(
            cls.__args__[0],
            key,
            registry,
            default=default,
            nullable=True,
            assign_widget=assign_widget,
            label=label,
            effects=effects,
            return_widget=return_widget,
            description=description,
            groups=groups,
            validators=validators,
        )

    if is_union(cls):
        variants = get_non_null_variants(cls)
        children = []
        converters = []
        for index, arg in enumerate(variants):
            child, converter = convert_child_to_childport(
                arg, registry, nullable=False, key="variant_" + str(index)
            )
            converters.append(converter)
            children.append(child)

        set_default = None
        if default:
            # We need to find the correct converter according
            # to the default value (checking the predicate)
            for index, child in enumerate(children):
                if predicate_port(child, default, registry):
                    set_default = converters[index](default)
                    break

        return PortInput(
            kind=PortKind.UNION,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            children=children,
            label=label,
            default=set_default,
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )

    if is_dict(cls):
        value_cls = get_dict_value_cls(cls)
        child, converter = convert_child_to_childport(
            value_cls, registry, nullable=False, key="..."
        )
        return PortInput(
            kind=PortKind.DICT,
            assignWidget=assign_widget,
            scope=PortScope.GLOBAL,
            returnWidget=return_widget,
            key=key,
            children=[child],
            label=label,
            default=(
                {key: converter(item) for key, item in default.items()}
                if default
                else None
            ),
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )

    if is_bool(cls) or (default is not None and isinstance(default, bool)):
        return PortInput(
            kind=PortKind.BOOL,
            scope=PortScope.GLOBAL,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            key=key,
            default=default,
            label=label,
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )  # catch bool is subclass of int

    if is_int(cls) or (default is not None and isinstance(default, int)):
        return PortInput(
            kind=PortKind.INT,
            assignWidget=assign_widget,
            scope=PortScope.GLOBAL,
            returnWidget=return_widget,
            key=key,
            default=default,
            label=label,
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )

    if is_float(cls) or (default is not None and isinstance(default, float)):
        return PortInput(
            kind=PortKind.FLOAT,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            default=default,
            label=label,
            nullable=nullable,
            validators=validators,
            effects=effects,
            description=description,
            groups=groups,
        )

    if is_datetime(cls) or (default is not None and isinstance(default, dt.datetime)):
        return PortInput(
            kind=PortKind.DATE,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            default=default,
            label=label,
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )

    if is_str(cls) or (default is not None and isinstance(default, str)):
        return PortInput(
            kind=PortKind.STRING,
            assignWidget=assign_widget,
            returnWidget=return_widget,
            scope=PortScope.GLOBAL,
            key=key,
            default=default,
            label=label,
            nullable=nullable,
            effects=effects,
            validators=validators,
            description=description,
            groups=groups,
        )

    if is_structure(cls):
        return registry.get_port_for_cls(
            cls,
            key,
            nullable=nullable,
            description=description,
            groups=groups,
            effects=effects,
            label=label,
            default=default,
            validators=validators,
            assign_widget=assign_widget,
            return_widget=return_widget,
        )

    raise NotImplementedError(f"Could not convert {cls} to a port")


GroupMap = Dict[str, List[str]]
AssignWidgetMap = Dict[str, List[AssignWidgetInput]]
ReturnWidgetMap = Dict[str, List[ReturnWidgetInput]]
EffectsMap = Dict[str, List[EffectInput]]


def snake_to_title_case(snake_str):
    # Split the string by underscores
    words = snake_str.split("_")

    # Capitalize each word
    capitalized_words = [word.capitalize() for word in words]

    # Join the words back into a single string with spaces in between
    title_case_str = " ".join(capitalized_words)

    return title_case_str


def prepare_definition(
    function: Callable,
    structure_registry: StructureRegistry,
    widgets: Optional[AssignWidgetMap] = None,
    return_widgets: Optional[ReturnWidgetMap] = None,
    groups: Optional[GroupMap] = None,
    effects: Optional[EffectsMap] = None,
    port_groups: List[PortGroupInput] = None,
    allow_empty_doc=True,
    collections: List[str] = [],
    interfaces: Optional[List[str]] = None,
    description: str = None,
    is_test_for: Optional[List[str]] = None,
    port_label_map: Optional[Dict[str, str]] = None,
    port_description_map: Optional[Dict[str, str]] = None,
    validators: Optional[Dict[str, List[ValidatorInput]]] = None,
    name: str = None,
    omitfirst=None,
    omitlast=None,
    omitkeys=[],
    allow_dev=True,
    allow_annotations: bool = True,
    **kwargs,  # additional kwargs can be ignored
) -> DefinitionInput:
    """Define

    Define a callable (async function, sync function, async generator, async
    generator) in the context of arkitekt and
    return its definition(input).

    Attention this definition is not yet registered in the
    arkitekt registry. This is done by the create_template function ( which will
    validate he definition with your local arkitekt instance
    and raise an error if the definition is not compatible with your arkitekt version)


    Args:
        function (): The function you want to define
    """

    assert structure_registry is not None, "You need to pass a StructureRegistry"

    is_generator = inspect.isasyncgenfunction(function) or inspect.isgeneratorfunction(
        function
    )

    sig = inspect.signature(function)
    widgets = widgets or {}
    effects = effects or {}
    validators = validators or {}

    port_groups = port_groups or []
    port_groups_name = [i.key for i in port_groups]
    groups = groups or {}
    for key, grouplist in groups.items():
        for group in grouplist:
            if group not in port_groups_name:
                raise DefinitionError(
                    f"Error mapping {group} to a group in port groups for port {key}:  Please define a PortGroup for {group}"
                )

    return_widgets = return_widgets or {}
    interfaces = interfaces or []
    collections = collections or []
    # Generate Args and Kwargs from the Annotation
    args: List[PortInput] = []
    returns: List[PortInput] = []

    # Docstring Parser to help with descriptions
    docstring = parse(function.__doc__)

    is_dev = False

    if not docstring.short_description and name is None:
        is_dev = True
        if not allow_dev:
            raise NonSufficientDocumentation(
                f"We are not in dev mode. Please provide a name or better document  {function.__name__}. Try docstring :)"
            )

    if not docstring.long_description and description is None and not allow_empty_doc:
        is_dev = True
        if not allow_dev:
            raise NonSufficientDocumentation(
                f"We are not in dev mode. Please provide a description or better document  {function.__name__}. Try docstring :)"
            )

    type_hints = get_type_hints(function, include_extras=allow_annotations)
    function_ins_annotation = sig.parameters

    doc_param_description_map = {
        param.arg_name: param.description for param in docstring.params
    }
    doc_param_label_map = {param.arg_name: param.arg_name for param in docstring.params}

    if docstring.many_returns:
        doc_param_description_map.update(
            {
                f"return{index}": param.description
                for index, param in enumerate(docstring.many_returns)
            }
        )
        doc_param_label_map.update(
            {
                f"return{index}": param.return_name
                for index, param in enumerate(docstring.many_returns)
            }
        )
    elif docstring.returns:
        doc_param_description_map.update({"return0": docstring.returns.description})
        doc_param_label_map.update({"return0": docstring.returns.return_name})

    if port_label_map:
        doc_param_label_map.update(port_label_map)
    if port_description_map:
        doc_param_description_map.update(port_description_map)

    for index, (key, value) in enumerate(function_ins_annotation.items()):
        # We can skip arguments if the builder is going to provide additional arguments
        if omitfirst is not None and index < omitfirst:
            continue
        if omitlast is not None and index > omitlast:
            continue
        if key in omitkeys:
            continue

        assign_widget = widgets.pop(key, None)
        port_effects = effects.pop(key, None)
        return_widget = return_widgets.pop(key, None)
        item_validators = validators.pop(key, None)
        default = value.default if value.default != inspect.Parameter.empty else None
        cls = type_hints.get(key, type(default) if default is not None else None)
        this_port_groups = groups.get(key, None)

        if cls is None:
            raise DefinitionError(
                f"Could not find type hint for {key} in {function.__name__}. Please provide a type hint (or default) for this argument."
            )

        try:
            args.append(
                convert_object_to_port(
                    cls,
                    key,
                    structure_registry,
                    assign_widget=assign_widget,
                    return_widget=return_widget,
                    default=default,
                    effects=port_effects,
                    nullable=value.default != inspect.Parameter.empty,
                    description=doc_param_description_map.pop(key, None),
                    label=doc_param_label_map.pop(key, None),
                    groups=this_port_groups,
                    validators=item_validators,
                )
            )
        except Exception as e:
            raise DefinitionError(
                f"Could not convert Argument of function {function.__name__} to"
                f" ArgPort: {value}"
            ) from e

    function_outs_annotation = sig.return_annotation

    if hasattr(function_outs_annotation, "_name"):
        if function_outs_annotation._name == "Tuple":
            try:
                for index, cls in enumerate(function_outs_annotation.__args__):
                    key = f"return{index}"
                    return_widget = return_widgets.pop(key, None)
                    assign_widget = widgets.pop(key, None)
                    port_effects = effects.pop(key, None)
                    this_port_groups = groups.pop(key, None)

                    returns.append(
                        convert_object_to_port(
                            cls,
                            key,
                            structure_registry,
                            return_widget=return_widget,
                            effects=port_effects,
                            description=doc_param_description_map.pop(key, None),
                            label=doc_param_label_map.pop(key, None),
                            assign_widget=assign_widget,
                            groups=this_port_groups,
                        )
                    )
            except Exception as e:
                raise DefinitionError(
                    f"Could not convert Return of function {function.__name__} to"
                    f" ArgPort: {cls}"
                ) from e
        else:
            try:
                key = "return0"
                return_widget = return_widgets.pop(key, None)
                assign_widget = widgets.pop(key, None)
                this_port_groups = groups.get(key, None)
                port_effects = effects.pop(key, None)
                returns.append(
                    convert_object_to_port(
                        function_outs_annotation,
                        key,
                        structure_registry,
                        return_widget=return_widget,
                        effects=port_effects,
                        description=doc_param_description_map.pop(key, None),
                        label=doc_param_label_map.pop(key, None),
                        assign_widget=assign_widget,
                        groups=this_port_groups,
                    )
                )  # Other types will be converted to normal lists and shit
            except Exception as e:
                raise DefinitionError(
                    f"Could not convert Return of function {function.__name__} to"
                    f" ArgPort: {function_outs_annotation}"
                ) from e
    else:
        # We are dealing with a non tuple return
        if function_outs_annotation is None:
            pass

        elif function_outs_annotation.__name__ != "_empty":  # Is it not empty
            key = "return0"
            return_widget = return_widgets.pop(key, None)
            assign_widget = widgets.pop(key, None)
            this_port_groups = groups.pop(key, None)
            port_effects = effects.pop(key, None)
            returns.append(
                convert_object_to_port(
                    function_outs_annotation,
                    "return0",
                    structure_registry,
                    assign_widget=assign_widget,
                    effects=port_effects,
                    description=doc_param_description_map.pop(key, None),
                    label=doc_param_label_map.pop(key, None),
                    return_widget=return_widget,
                    groups=this_port_groups,
                )
            )

    # Documentation Parsing

    name = name or docstring.short_description or snake_to_title_case(function.__name__)
    description = description or docstring.long_description or "No Description"

    if widgets:
        raise DefinitionError(
            f"Could not find the following ports for the widgets in the function {function.__name__}: {','.join(widgets.keys())}. Did you forget the type hint?"
        )
    if return_widgets:
        raise DefinitionError(
            f"Could not find the following ports for the return widgets in the function {function.__name__}: {','.join(return_widgets.keys())}. Did you forget the type hint?"
        )
    if port_label_map:
        raise DefinitionError(
            f"Could not find the following ports for the labels in the function {function.__name__}: {','.join(port_label_map.keys())}. Did you forget the type hint?"
        )
    if port_description_map:
        raise DefinitionError(
            f"Could not find the following ports for the descriptions in the function {function.__name__}: {','.join(port_description_map.keys())}. Did you forget the type hint?"
        )

    x = DefinitionInput(
        **{
            "name": name,
            "description": description,
            "collections": collections,
            "args": args,
            "returns": returns,
            "kind": NodeKind.GENERATOR if is_generator else NodeKind.FUNCTION,
            "interfaces": interfaces,
            "portGroups": port_groups,
            "isDev": is_dev,
            "isTestFor": is_test_for or [],
        }
    )

    return x



class AgentException(Exception):
    """
    Base class for all exceptions raised by the Agent.
    """


class ProvisionException(AgentException):
    """
    Base class for all exceptions raised by the Agent.
    """


class ExtensionError(AgentException):
    """
    Base class for all exceptions raised by an Extension of the Agent.
    """

class HookError(AgentException):
    """
    Base class for all exceptions raised by a hook
    """


class StartupHookError(HookError):
    """
    Raised when a startup hook fails
    """


class BackgroundHookError(HookError):
    """
    Raised when a background hook fails
    """


from rekuest_next.api.schema import DependencyInput


def declare(function_or_hash, optional=False, **kwargs):
    """Declare a function that needs to be
    called outside of your application.

    Args:
        function_or_hash (str or callable): The function or hash that needs to be declared.
        optional (bool, optional): Whether the dependency is optional. Defaults to False.

    Returns:
        DependencyInput: The dependency input object.

    """

    if callable(function_or_hash):
        assert hasattr(
            function_or_hash, "__definition_hash__"
        ), "Only previously registered function can be declared. If you simply want to call your function locally, just call it. This is a feature to declare a function that might call outside of your application"

        return DependencyInput(
            hash=function_or_hash.__definition_hash__, optional=optional
        )

    else:
        assert isinstance(
            function_or_hash, str
        ), "Only hash or function can be declared"
        return DependencyInput(hash=function_or_hash, optional=optional)



from typing import Any, List, Optional, TypeVar, Literal, Union, Dict
from pydantic import BaseModel
from rekuest_next.api.schema import (
    AssignationEventKind,
    ProvisionEventKind,
)
from enum import Enum
from pydantic import Field
import uuid


class MessageType(str, Enum):
    ASSIGN = "ASSIGN"
    CANCEL = "CANCEL"
    INTERRUPT = "INTERRUPT"
    PROVIDE = "PROVIDE"
    UNPROVIDE = "UNPROVIDE"
    ASSIGNATION_EVENT = "ASSIGNATION_EVENT"
    PROVISION_EVENT = "PROVISION_EVENT"
    INIT = "INIT"
    HEARTBEAT = "HEARTBEAT"


class Message(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    type: MessageType


class Assign(Message):
    type: Literal[MessageType.ASSIGN] = MessageType.ASSIGN
    assignation: str
    reference: Optional[str]
    provision: Optional[str]
    reservation: Optional[str]
    args: Optional[Dict[str, Any]]
    message: Optional[str]
    user: Optional[str]


class Cancel(Message):
    type: Literal[MessageType.CANCEL] = MessageType.CANCEL
    assignation: str
    provision: str


class Interrupt(Message):
    type: Literal[MessageType.INTERRUPT] = MessageType.INTERRUPT
    assignation: str
    provision: str


class Provide(Message):
    type: Literal[MessageType.PROVIDE] = MessageType.PROVIDE
    provision: str


class Unprovide(Message):
    type: Literal[MessageType.UNPROVIDE] = MessageType.UNPROVIDE
    provision: str
    message: Optional[str]


class AssignationEvent(Message):
    type: Literal[MessageType.ASSIGNATION_EVENT] = MessageType.ASSIGNATION_EVENT
    assignation: str
    kind: AssignationEventKind
    message: Optional[str]
    returns: Optional[Dict[str, Any]] = Field(default_factory=None)
    persist: Optional[bool]
    progress: Optional[int]
    log: Optional[bool]
    status: Optional[AssignationEventKind]


class ProvisionEvent(Message):
    type: Literal[MessageType.PROVISION_EVENT] = MessageType.PROVISION_EVENT
    provision: str
    kind: ProvisionEventKind
    message: Optional[str]
    user: Optional[str]


class AssignInquiry(BaseModel):
    assignation: str


class ProvideInquiry(BaseModel):
    provision: str


class Init(Message):
    type: Literal[MessageType.INIT] = MessageType.INIT
    instance_id: str = None
    agent: str = None
    registry: str = None
    provisions: list[Provide] = []
    inquiries: list[AssignInquiry] = []


InMessage = Union[Init, Assign, Cancel, Interrupt, Provide, Unprovide]
OutMessage = Union[AssignationEvent, ProvisionEvent]



from typing import (
    List,
    Optional,
    Dict,
    Any,
    Protocol,
    runtime_checkable,
)

from rekuest_next.messages import Assignation, Reservation, Unassignation, Unreservation
from rekuest_next.api.schema import (
    AssignationStatus,
    ReservationStatus,
    ReserveParamsInput,
)


@runtime_checkable
class PostmanTransport(Protocol):
    connected = False

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        ...

    async def aunassign(self, assignation: str) -> Unassignation:
        ...

    async def areserve(
        self,
        node: str,
        params: ReserveParamsInput = None,
        provision: str = None,
        reference: str = "default",
    ) -> Reservation:
        ...

    async def aunreserve(
        self,
        reservation: str,
    ) -> Unreservation:
        ...

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        ...

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        ...



import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Awaitable, Callable
from koil.helpers import iterate_spawned, run_spawned
from pydantic import BaseModel, Field
from rekuest_next.actors.base import SerializingActor
from rekuest_next.messages import Assign, Provide
from rekuest_next.api.schema import AssignationEventKind
from rekuest_next.structures.serialization.actor import expand_inputs, shrink_outputs
from rekuest_next.actors.contexts import AssignmentContext
from rekuest_next.actors.types import OnProvide, OnUnprovide, Assignment, Unassignment
from rekuest_next.collection.collector import Collector
from rekuest_next.actors.transport.types import AssignTransport
from rekuest_next.structures.parse_collectables import parse_collectable
from rekuest_next.structures.errors import SerializationError

logger = logging.getLogger(__name__)


async def async_none_provide(prov: Provide):
    """Do nothing on provide"""
    return None


async def async_none_unprovide():
    """Do nothing on unprovide"""
    return None


class FunctionalActor(BaseModel):
    assign: Callable[..., Any]
    on_provide: OnProvide = Field(default=async_none_provide)
    on_unprovide: OnUnprovide = Field(default=async_none_unprovide)

    class Config:
        arbitrary_types_allowed = True


class AsyncFuncActor(SerializingActor):
    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            print(assignment.args)
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )

            print(params)

            await transport.log_event(
                kind=AssignationEventKind.ASSIGN,
                message="Assigned to actor",
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                returns = await self.assign(**params)

            returns = await shrink_outputs(
                self.definition,
                returns,
                structure_registry=self.structure_registry,
                skip_shrinking=not self.shrink_outputs,
            )

            collector.register(assignment, parse_collectable(self.definition, returns))

            await transport.log_event(
                kind=AssignationEventKind.YIELD,
                returns=returns,
            )

            await transport.log_event(
                kind=AssignationEventKind.DONE,
            )

        except SerializationError as ex:
            logger.critical("Assignation error", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.ERROR,
                message=str(ex),
            )

        except AssertionError as ex:
            logger.critical("Assignation error", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except Exception as e:
            logger.critical("Assignation error", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=repr(e),
            )


class AsyncGenActor(SerializingActor):
    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )

            await transport.log_event(
                kind=AssignationEventKind.ASSIGN,
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                async for returns in self.assign(**params):
                    returns = await shrink_outputs(
                        self.definition,
                        returns,
                        structure_registry=self.structure_registry,
                        skip_shrinking=not self.shrink_outputs,
                    )

                    collector.register(
                        assignment, parse_collectable(self.definition, returns)
                    )

                    await transport.log_event(
                        kind=AssignationEventKind.YIELD,
                        returns=returns,
                    )

            await transport.log_event(kind=AssignationEventKind.DONE)

        except SerializationError as ex:
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except AssertionError as ex:
            await transport.log_event(
                kind=AssignationEventKind.ERROR,
                message=str(ex),
            )

        except Exception as ex:
            logger.critical("Error in actor", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

            raise ex


class FunctionalFuncActor(FunctionalActor, AsyncFuncActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True


class FunctionalGenActor(FunctionalActor, AsyncGenActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True


class ThreadedFuncActor(SerializingActor):
    executor: ThreadPoolExecutor = Field(default_factory=lambda: ThreadPoolExecutor(1))

    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            logger.info("Assigning Number two")
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )

            await transport.log_event(
                kind=AssignationEventKind.ASSIGN,
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                returns = await run_spawned(
                    self.assign, **params, executor=self.executor, pass_context=True
                )

            returns = await shrink_outputs(
                self.definition,
                returns,
                structure_registry=self.structure_registry,
                skip_shrinking=not self.shrink_outputs,
            )

            collector.register(assignment, parse_collectable(self.definition, returns))

            await transport.log_event(
                kind=AssignationEventKind.YIELD,
                returns=returns,
            )

            await transport.log_event(
                kind=AssignationEventKind.DONE,
            )

        except SerializationError as ex:
            logger.error("Serializing Error in actor", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except AssertionError as ex:
            logger.error("AssertionError in actor", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(e),
            )


class ThreadedGenActor(SerializingActor):
    executor: ThreadPoolExecutor = Field(default_factory=lambda: ThreadPoolExecutor(4))

    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )
            await transport.log_event(
                kind=AssignationEventKind.ASSIGN,
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                async for returns in iterate_spawned(
                    self.assign, **params, executor=self.executor, pass_context=True
                ):
                    returns = await shrink_outputs(
                        self.definition,
                        returns,
                        structure_registry=self.structure_registry,
                        skip_shrinking=not self.shrink_outputs,
                    )

                    collector.register(
                        assignment, parse_collectable(self.definition, returns)
                    )

                    await transport.log_event(
                        kind=AssignationEventKind.YIELD,
                        returns=returns,
                    )

            await transport.log_event(kind=AssignationEventKind.DONE)

        except AssertionError as ex:
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except SerializationError as ex:
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except Exception as e:
            logging.critical(f"Assignation Error {assignment} {e}", exc_info=True)
            await transport.log_event(
                kind=AssignationEventKind.CRITICAL,
                message=str(e),
            )

            raise e


class ProcessedGenActor(SerializingActor):
    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )
            await transport.change(
                status=AssignationEventKind.ASSIGN,
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                async for returns in iterate_processed(self.assign, **params):
                    returns = await shrink_outputs(
                        self.definition,
                        returns,
                        structure_registry=self.structure_registry,
                        skip_shrinking=not self.shrink_outputs,
                    )

                    collector.register(
                        assignment, parse_collectable(self.definition, returns)
                    )

                    await self.transport.change(
                        status=AssignationEventKind.ASSIGN,
                        returns=returns,
                    )

            await transport.change(status=AssignationEventKind.DONE)

        except AssertionError as ex:
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except SerializationError as ex:
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except Exception as e:
            logging.critical(f"Assignation Error {assignment} {e}", exc_info=True)
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(e),
            )

            raise e


class ProcessedFuncActor(SerializingActor):
    async def on_assign(
        self,
        assignment: Assignment,
        collector: Collector,
        transport: AssignTransport,
    ):
        try:
            logger.info("Assigning Number two")
            params = await expand_inputs(
                self.definition,
                assignment.args,
                structure_registry=self.structure_registry,
                skip_expanding=not self.expand_inputs,
            )

            await transport.change(
                status=AssignationEventKind.ASSIGN,
            )

            async with AssignmentContext(
                assignment=assignment, transport=transport, passport=self.passport
            ):
                returns = await run_processed(
                    self.assign,
                    **params,
                )

            returns = await shrink_outputs(
                self.definition,
                returns,
                structure_registry=self.structure_registry,
                skip_shrinking=not self.shrink_outputs,
            )

            collector.register(assignment, parse_collectable(self.definition, returns))

            await transport.change(
                status=AssignationEventKind.RETURNED,
                returns=returns,
            )

        except SerializationError as ex:
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except AssertionError as ex:
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(ex),
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await transport.change(
                status=AssignationEventKind.CRITICAL,
                message=str(e),
            )


class FunctionalThreadedFuncActor(FunctionalActor, ThreadedFuncActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True


class FunctionalThreadedGenActor(FunctionalActor, ThreadedGenActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True


class FunctionalProcessedFuncActor(FunctionalActor, ProcessedFuncActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True


class FunctionalProcessedGenActor(FunctionalActor, ProcessedGenActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    class Config:
        arbitrary_types_allowed = True



from rekuest_next.scalars import (
    InstanceId,
    ValidatorFunction,
    NodeHash,
    Identifier,
    SearchQuery,
    Args,
)
from typing_extensions import Literal
from typing import AsyncIterator, Any, Tuple, Iterator, List, Optional
from rekuest_next.funcs import aexecute, asubscribe, execute, subscribe
from enum import Enum
from pydantic import Field, BaseModel
from rekuest_next.traits.ports import PortTrait, ReturnWidgetInputTrait
from rekuest_next.rath import RekuestNextRath
from rekuest_next.traits.node import Reserve
from datetime import datetime
from rath.scalars import ID


class AssignWidgetKind(str, Enum):
    SEARCH = "SEARCH"
    CHOICE = "CHOICE"
    SLIDER = "SLIDER"
    CUSTOM = "CUSTOM"
    STRING = "STRING"


class ReturnWidgetKind(str, Enum):
    CHOICE = "CHOICE"
    CUSTOM = "CUSTOM"


class LogicalCondition(str, Enum):
    IS = "IS"
    IS_NOT = "IS_NOT"
    IN = "IN"


class EffectKind(str, Enum):
    MESSAGE = "MESSAGE"
    CUSTOM = "CUSTOM"


class NodeKind(str, Enum):
    FUNCTION = "FUNCTION"
    GENERATOR = "GENERATOR"


class PortKind(str, Enum):
    INT = "INT"
    STRING = "STRING"
    STRUCTURE = "STRUCTURE"
    LIST = "LIST"
    BOOL = "BOOL"
    DICT = "DICT"
    FLOAT = "FLOAT"
    DATE = "DATE"
    UNION = "UNION"
    MODEL = "MODEL"


class ReservationEventKind(str, Enum):
    PENDING = "PENDING"
    CREATE = "CREATE"
    RESCHEDULE = "RESCHEDULE"
    DELETED = "DELETED"
    CHANGE = "CHANGE"
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    UNCONNECTED = "UNCONNECTED"
    ENDED = "ENDED"
    UNHAPPY = "UNHAPPY"
    HAPPY = "HAPPY"
    LOG = "LOG"


class ProvisionEventKind(str, Enum):
    CHANGE = "CHANGE"
    UNHAPPY = "UNHAPPY"
    PENDING = "PENDING"
    CRITICAL = "CRITICAL"
    DENIED = "DENIED"
    ACTIVE = "ACTIVE"
    REFUSED = "REFUSED"
    INACTIVE = "INACTIVE"
    CANCELING = "CANCELING"
    DISCONNECTED = "DISCONNECTED"
    RECONNECTING = "RECONNECTING"
    ERROR = "ERROR"
    ENDED = "ENDED"
    CANCELLED = "CANCELLED"
    BOUND = "BOUND"
    PROVIDING = "PROVIDING"
    LOG = "LOG"


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    ERROR = "ERROR"
    WARN = "WARN"
    CRITICAL = "CRITICAL"


class PortScope(str, Enum):
    GLOBAL = "GLOBAL"
    LOCAL = "LOCAL"


class AssignationEventKind(str, Enum):
    BOUND = "BOUND"
    ASSIGN = "ASSIGN"
    PROGRESS = "PROGRESS"
    DISCONNECTED = "DISCONNECTED"
    YIELD = "YIELD"
    DONE = "DONE"
    LOG = "LOG"
    CANCELING = "CANCELING"
    CANCELLED = "CANCELLED"
    INTERUPTING = "INTERUPTING"
    INTERUPTED = "INTERUPTED"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class HookKind(str, Enum):
    CLEANUP = "CLEANUP"
    INIT = "INIT"


class CreateTemplateInput(BaseModel):
    template: "TemplateInput"
    instance_id: InstanceId = Field(alias="instanceId")
    extension: str

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class TemplateInput(BaseModel):
    definition: "DefinitionInput"
    dependencies: Tuple["DependencyInput", ...]
    interface: str
    params: Optional[Any]
    dynamic: bool
    logo: Optional[str]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class DefinitionInput(BaseModel):
    description: Optional[str]
    collections: Tuple[str, ...]
    name: str
    port_groups: Tuple["PortGroupInput", ...] = Field(alias="portGroups")
    args: Tuple["PortInput", ...]
    returns: Tuple["PortInput", ...]
    kind: NodeKind
    is_test_for: Tuple[str, ...] = Field(alias="isTestFor")
    interfaces: Tuple[str, ...]
    is_dev: bool = Field(alias="isDev")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class PortGroupInput(BaseModel):
    key: str
    hidden: bool

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class PortInput(PortTrait, BaseModel):
    validators: Optional[Tuple["ValidatorInput", ...]]
    key: str
    scope: PortScope
    label: Optional[str]
    kind: PortKind
    description: Optional[str]
    identifier: Optional[str]
    nullable: bool
    effects: Optional[Tuple["EffectInput", ...]]
    default: Optional[Any]
    children: Optional[Tuple["ChildPortInput", ...]]
    assign_widget: Optional["AssignWidgetInput"] = Field(alias="assignWidget")
    return_widget: Optional["ReturnWidgetInput"] = Field(alias="returnWidget")
    groups: Optional[Tuple[str, ...]]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class ValidatorInput(BaseModel):
    function: ValidatorFunction
    dependencies: Optional[Tuple[str, ...]]
    label: Optional[str]
    error_message: Optional[str] = Field(alias="errorMessage")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class EffectInput(BaseModel):
    label: str
    description: Optional[str]
    dependencies: Tuple["EffectDependencyInput", ...]
    kind: EffectKind

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class EffectDependencyInput(BaseModel):
    key: str
    condition: LogicalCondition
    value: Any

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class ChildPortInput(PortTrait, BaseModel):
    default: Optional[Any]
    key: str
    label: Optional[str]
    kind: PortKind
    scope: PortScope
    description: Optional[str]
    identifier: Optional[Identifier]
    nullable: bool
    children: Optional[Tuple["ChildPortInput", ...]]
    effects: Optional[Tuple[EffectInput, ...]]
    assign_widget: Optional["AssignWidgetInput"] = Field(alias="assignWidget")
    return_widget: Optional["ReturnWidgetInput"] = Field(alias="returnWidget")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class AssignWidgetInput(BaseModel):
    as_paragraph: Optional[bool] = Field(alias="asParagraph")
    kind: AssignWidgetKind
    query: Optional[SearchQuery]
    choices: Optional[Tuple["ChoiceInput", ...]]
    min: Optional[int]
    max: Optional[int]
    step: Optional[int]
    placeholder: Optional[str]
    hook: Optional[str]
    ward: Optional[str]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class ChoiceInput(BaseModel):
    value: Any
    label: str
    description: Optional[str]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class ReturnWidgetInput(ReturnWidgetInputTrait, BaseModel):
    kind: ReturnWidgetKind
    query: Optional[SearchQuery]
    choices: Optional[Tuple[ChoiceInput, ...]]
    min: Optional[int]
    max: Optional[int]
    step: Optional[int]
    placeholder: Optional[str]
    hook: Optional[str]
    ward: Optional[str]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class DependencyInput(BaseModel):
    hash: Optional[NodeHash]
    reference: Optional[str]
    binds: Optional["BindsInput"]
    optional: bool
    viable_instances: Optional[int] = Field(alias="viableInstances")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class BindsInput(BaseModel):
    templates: Optional[Tuple[str, ...]]
    clients: Optional[Tuple[str, ...]]
    desired_instances: int = Field(alias="desiredInstances")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class SetExtensionTemplatesInput(BaseModel):
    templates: Tuple[TemplateInput, ...]
    instance_id: InstanceId = Field(alias="instanceId")
    extension: str
    run_cleanup: bool = Field(alias="runCleanup")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class AssignInput(BaseModel):
    instance_id: InstanceId = Field(alias="instanceId")
    node: Optional[ID]
    template: Optional[ID]
    reservation: Optional[ID]
    hooks: Optional[Tuple["HookInput", ...]]
    args: Args
    reference: Optional[str]
    parent: Optional[ID]
    cached: bool
    log: bool
    is_hook: bool = Field(alias="isHook")

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class HookInput(BaseModel):
    kind: HookKind
    hash: str

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class CancelInput(BaseModel):
    assignation: ID

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class InterruptInput(BaseModel):
    assignation: ID

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class ReserveInput(BaseModel):
    assignation_id: Optional[str] = Field(alias="assignationId")
    instance_id: InstanceId = Field(alias="instanceId")
    node: Optional[ID]
    template: Optional[ID]
    title: Optional[str]
    hash: Optional[NodeHash]
    reference: Optional[str]
    binds: Optional[BindsInput]

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class UnreserveInput(BaseModel):
    reservation: ID

    class Config:
        """A config class"""

        frozen = True
        extra = "forbid"
        use_enum_values = True


class TestCaseFragmentNode(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class TestCaseFragment(BaseModel):
    typename: Optional[Literal["TestCase"]] = Field(alias="__typename", exclude=True)
    id: ID
    node: TestCaseFragmentNode
    key: str
    is_benchmark: bool = Field(alias="isBenchmark")
    description: str
    name: str

    class Config:
        """A config class"""

        frozen = True


class TestResultFragmentCase(BaseModel):
    typename: Optional[Literal["TestCase"]] = Field(alias="__typename", exclude=True)
    id: ID
    key: str

    class Config:
        """A config class"""

        frozen = True


class TestResultFragment(BaseModel):
    typename: Optional[Literal["TestResult"]] = Field(alias="__typename", exclude=True)
    id: ID
    case: TestResultFragmentCase
    passed: bool

    class Config:
        """A config class"""

        frozen = True


class ProvisionFragment(BaseModel):
    typename: Optional[Literal["Provision"]] = Field(alias="__typename", exclude=True)
    id: ID
    status: ProvisionEventKind
    template: "TemplateFragment"

    class Config:
        """A config class"""

        frozen = True


class ChildPortNestedFragmentChildren(PortTrait, BaseModel):
    typename: Optional[Literal["ChildPort"]] = Field(alias="__typename", exclude=True)
    identifier: Optional[Identifier]
    nullable: bool
    kind: PortKind

    class Config:
        """A config class"""

        frozen = True


class ChildPortNestedFragment(PortTrait, BaseModel):
    typename: Optional[Literal["ChildPort"]] = Field(alias="__typename", exclude=True)
    key: str
    kind: PortKind
    children: Optional[Tuple[ChildPortNestedFragmentChildren, ...]]
    identifier: Optional[Identifier]
    nullable: bool

    class Config:
        """A config class"""

        frozen = True


class ChildPortFragment(PortTrait, BaseModel):
    typename: Optional[Literal["ChildPort"]] = Field(alias="__typename", exclude=True)
    key: str
    kind: PortKind
    identifier: Optional[Identifier]
    children: Optional[Tuple[ChildPortNestedFragment, ...]]
    nullable: bool

    class Config:
        """A config class"""

        frozen = True


class PortFragmentValidators(BaseModel):
    typename: Optional[Literal["Validator"]] = Field(alias="__typename", exclude=True)
    function: ValidatorFunction
    error_message: Optional[str] = Field(alias="errorMessage")
    dependencies: Optional[Tuple[str, ...]]
    label: Optional[str]

    class Config:
        """A config class"""

        frozen = True


class PortFragment(PortTrait, BaseModel):
    typename: Optional[Literal["Port"]] = Field(alias="__typename", exclude=True)
    key: str
    label: Optional[str]
    nullable: bool
    description: Optional[str]
    default: Optional[Any]
    kind: PortKind
    identifier: Optional[Identifier]
    children: Optional[Tuple[ChildPortFragment, ...]]
    validators: Optional[Tuple[PortFragmentValidators, ...]]

    class Config:
        """A config class"""

        frozen = True


class AgentFragmentRegistryApp(BaseModel):
    typename: Optional[Literal["App"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class AgentFragmentRegistryUser(BaseModel):
    typename: Optional[Literal["User"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class AgentFragmentRegistry(BaseModel):
    typename: Optional[Literal["Registry"]] = Field(alias="__typename", exclude=True)
    app: AgentFragmentRegistryApp
    user: AgentFragmentRegistryUser

    class Config:
        """A config class"""

        frozen = True


class AgentFragment(BaseModel):
    typename: Optional[Literal["Agent"]] = Field(alias="__typename", exclude=True)
    registry: AgentFragmentRegistry

    class Config:
        """A config class"""

        frozen = True


class ReservationFragmentNode(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    id: ID
    hash: NodeHash

    class Config:
        """A config class"""

        frozen = True


class ReservationFragmentWaiter(BaseModel):
    typename: Optional[Literal["Waiter"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class ReservationFragment(BaseModel):
    typename: Optional[Literal["Reservation"]] = Field(alias="__typename", exclude=True)
    id: ID
    status: ReservationEventKind
    node: ReservationFragmentNode
    waiter: ReservationFragmentWaiter
    reference: str
    updated_at: datetime = Field(alias="updatedAt")

    class Config:
        """A config class"""

        frozen = True


class AssignationFragmentParent(BaseModel):
    typename: Optional[Literal["Assignation"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class AssignationFragmentEvents(BaseModel):
    typename: Optional[Literal["AssignationEvent"]] = Field(
        alias="__typename", exclude=True
    )
    id: ID
    returns: Optional[Any]
    level: Optional[LogLevel]

    class Config:
        """A config class"""

        frozen = True


class AssignationFragment(BaseModel):
    typename: Optional[Literal["Assignation"]] = Field(alias="__typename", exclude=True)
    args: Any
    id: ID
    parent: Optional[AssignationFragmentParent]
    id: ID
    status: AssignationEventKind
    events: Tuple[AssignationFragmentEvents, ...]
    reference: Optional[str]
    updated_at: datetime = Field(alias="updatedAt")

    class Config:
        """A config class"""

        frozen = True


class AssignationEventFragment(BaseModel):
    typename: Optional[Literal["AssignationEvent"]] = Field(
        alias="__typename", exclude=True
    )
    id: ID
    kind: AssignationEventKind
    returns: Optional[Any]
    reference: str
    message: Optional[str]
    progress: Optional[int]

    class Config:
        """A config class"""

        frozen = True


class AssignationChangeEventFragment(BaseModel):
    typename: Optional[Literal["AssignationChangeEvent"]] = Field(
        alias="__typename", exclude=True
    )
    create: Optional[AssignationFragment]
    event: Optional[AssignationEventFragment]

    class Config:
        """A config class"""

        frozen = True


class TemplateFragmentAgentRegistry(BaseModel):
    typename: Optional[Literal["Registry"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class TemplateFragmentAgent(BaseModel):
    typename: Optional[Literal["Agent"]] = Field(alias="__typename", exclude=True)
    registry: TemplateFragmentAgentRegistry

    class Config:
        """A config class"""

        frozen = True


class TemplateFragment(BaseModel):
    typename: Optional[Literal["Template"]] = Field(alias="__typename", exclude=True)
    id: ID
    agent: TemplateFragmentAgent
    node: "NodeFragment"
    params: Any
    extension: str
    interface: str

    class Config:
        """A config class"""

        frozen = True


class DefinitionFragmentCollections(BaseModel):
    typename: Optional[Literal["Collection"]] = Field(alias="__typename", exclude=True)
    name: str

    class Config:
        """A config class"""

        frozen = True


class DefinitionFragmentIstestfor(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class DefinitionFragmentPortgroups(BaseModel):
    typename: Optional[Literal["PortGroup"]] = Field(alias="__typename", exclude=True)
    key: str

    class Config:
        """A config class"""

        frozen = True


class DefinitionFragment(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    args: Tuple[PortFragment, ...]
    returns: Tuple[PortFragment, ...]
    kind: NodeKind
    name: str
    description: Optional[str]
    collections: Tuple[DefinitionFragmentCollections, ...]
    is_dev: bool = Field(alias="isDev")
    is_test_for: Tuple[DefinitionFragmentIstestfor, ...] = Field(alias="isTestFor")
    port_groups: Tuple[DefinitionFragmentPortgroups, ...] = Field(alias="portGroups")

    class Config:
        """A config class"""

        frozen = True


class NodeFragment(DefinitionFragment, Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    hash: NodeHash
    id: ID

    class Config:
        """A config class"""

        frozen = True


class Create_testcaseMutation(BaseModel):
    create_test_case: TestCaseFragment = Field(alias="createTestCase")

    class Arguments(BaseModel):
        node: ID
        key: str
        is_benchmark: Optional[bool] = Field(default=None)
        description: str
        name: str

    class Meta:
        document = "fragment TestCase on TestCase {\n  id\n  node {\n    id\n  }\n  key\n  isBenchmark\n  description\n  name\n}\n\nmutation create_testcase($node: ID!, $key: String!, $is_benchmark: Boolean, $description: String!, $name: String!) {\n  createTestCase(\n    input: {node: $node, key: $key, isBenchmark: $is_benchmark, description: $description, name: $name}\n  ) {\n    ...TestCase\n  }\n}"


class Create_testresultMutation(BaseModel):
    create_test_result: TestResultFragment = Field(alias="createTestResult")

    class Arguments(BaseModel):
        case: ID
        template: ID
        passed: bool
        result: Optional[str] = Field(default=None)

    class Meta:
        document = "fragment TestResult on TestResult {\n  id\n  case {\n    id\n    key\n  }\n  passed\n}\n\nmutation create_testresult($case: ID!, $template: ID!, $passed: Boolean!, $result: String) {\n  createTestResult(\n    input: {case: $case, template: $template, passed: $passed, result: $result}\n  ) {\n    ...TestResult\n  }\n}"


class EnsureAgentMutationEnsureagent(BaseModel):
    typename: Optional[Literal["Agent"]] = Field(alias="__typename", exclude=True)
    id: ID
    instance_id: InstanceId = Field(alias="instanceId")
    extensions: Tuple[str, ...]

    class Config:
        """A config class"""

        frozen = True


class EnsureAgentMutation(BaseModel):
    ensure_agent: EnsureAgentMutationEnsureagent = Field(alias="ensureAgent")

    class Arguments(BaseModel):
        instance_id: InstanceId = Field(alias="instanceId")
        extensions: Optional[List[str]] = Field(default=None)

    class Meta:
        document = "mutation EnsureAgent($instanceId: InstanceId!, $extensions: [String!]) {\n  ensureAgent(input: {instanceId: $instanceId, extensions: $extensions}) {\n    id\n    instanceId\n    extensions\n  }\n}"


class ReserveMutation(BaseModel):
    reserve: ReservationFragment

    class Arguments(BaseModel):
        input: ReserveInput

    class Meta:
        document = "fragment Reservation on Reservation {\n  id\n  status\n  node {\n    id\n    hash\n  }\n  waiter {\n    id\n  }\n  reference\n  updatedAt\n}\n\nmutation reserve($input: ReserveInput!) {\n  reserve(input: $input) {\n    ...Reservation\n  }\n}"


class UnreserveMutation(BaseModel):
    unreserve: str

    class Arguments(BaseModel):
        input: UnreserveInput

    class Meta:
        document = "mutation unreserve($input: UnreserveInput!) {\n  unreserve(input: $input)\n}"


class AssignMutation(BaseModel):
    assign: AssignationFragment

    class Arguments(BaseModel):
        input: AssignInput

    class Meta:
        document = "fragment Assignation on Assignation {\n  args\n  id\n  parent {\n    id\n  }\n  id\n  status\n  events {\n    id\n    returns\n    level\n  }\n  reference\n  updatedAt\n}\n\nmutation assign($input: AssignInput!) {\n  assign(input: $input) {\n    ...Assignation\n  }\n}"


class CancelMutation(BaseModel):
    cancel: AssignationFragment

    class Arguments(BaseModel):
        input: CancelInput

    class Meta:
        document = "fragment Assignation on Assignation {\n  args\n  id\n  parent {\n    id\n  }\n  id\n  status\n  events {\n    id\n    returns\n    level\n  }\n  reference\n  updatedAt\n}\n\nmutation cancel($input: CancelInput!) {\n  cancel(input: $input) {\n    ...Assignation\n  }\n}"


class InterruptMutation(BaseModel):
    interrupt: AssignationFragment

    class Arguments(BaseModel):
        input: InterruptInput

    class Meta:
        document = "fragment Assignation on Assignation {\n  args\n  id\n  parent {\n    id\n  }\n  id\n  status\n  events {\n    id\n    returns\n    level\n  }\n  reference\n  updatedAt\n}\n\nmutation interrupt($input: InterruptInput!) {\n  interrupt(input: $input) {\n    ...Assignation\n  }\n}"


class CreateHardwareRecordMutationCreatehardwarerecordAgent(BaseModel):
    typename: Optional[Literal["Agent"]] = Field(alias="__typename", exclude=True)
    id: ID

    class Config:
        """A config class"""

        frozen = True


class CreateHardwareRecordMutationCreatehardwarerecord(BaseModel):
    typename: Optional[Literal["HardwareRecord"]] = Field(
        alias="__typename", exclude=True
    )
    id: ID
    cpu_count: int = Field(alias="cpuCount")
    agent: CreateHardwareRecordMutationCreatehardwarerecordAgent

    class Config:
        """A config class"""

        frozen = True


class CreateHardwareRecordMutation(BaseModel):
    create_hardware_record: CreateHardwareRecordMutationCreatehardwarerecord = Field(
        alias="createHardwareRecord"
    )

    class Arguments(BaseModel):
        cpu_count: Optional[int] = Field(alias="cpuCount", default=None)
        cpu_frequency: Optional[float] = Field(alias="cpuFrequency", default=None)
        cpu_vendor_name: Optional[str] = Field(alias="cpuVendorName", default=None)

    class Meta:
        document = "mutation CreateHardwareRecord($cpuCount: Int, $cpuFrequency: Float, $cpuVendorName: String) {\n  createHardwareRecord(\n    input: {cpuCount: $cpuCount, cpuFrequency: $cpuFrequency, cpuVendorName: $cpuVendorName}\n  ) {\n    id\n    cpuCount\n    agent {\n      id\n    }\n  }\n}"


class CreateTemplateMutation(BaseModel):
    create_template: TemplateFragment = Field(alias="createTemplate")

    class Arguments(BaseModel):
        input: CreateTemplateInput

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nfragment Template on Template {\n  id\n  agent {\n    registry {\n      id\n    }\n  }\n  node {\n    ...Node\n  }\n  params\n  extension\n  interface\n}\n\nmutation createTemplate($input: CreateTemplateInput!) {\n  createTemplate(input: $input) {\n    ...Template\n  }\n}"


class SetExtensionTemplatesMutation(BaseModel):
    set_extension_templates: Tuple[TemplateFragment, ...] = Field(
        alias="setExtensionTemplates"
    )

    class Arguments(BaseModel):
        input: SetExtensionTemplatesInput

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nfragment Template on Template {\n  id\n  agent {\n    registry {\n      id\n    }\n  }\n  node {\n    ...Node\n  }\n  params\n  extension\n  interface\n}\n\nmutation SetExtensionTemplates($input: SetExtensionTemplatesInput!) {\n  setExtensionTemplates(input: $input) {\n    ...Template\n  }\n}"


class WatchReservationsSubscription(BaseModel):
    reservations: ReservationFragment

    class Arguments(BaseModel):
        instance_id: InstanceId = Field(alias="instanceId")

    class Meta:
        document = "fragment Reservation on Reservation {\n  id\n  status\n  node {\n    id\n    hash\n  }\n  waiter {\n    id\n  }\n  reference\n  updatedAt\n}\n\nsubscription WatchReservations($instanceId: InstanceId!) {\n  reservations(instanceId: $instanceId) {\n    ...Reservation\n  }\n}"


class WatchAssignationsSubscription(BaseModel):
    assignations: AssignationChangeEventFragment

    class Arguments(BaseModel):
        instance_id: InstanceId = Field(alias="instanceId")

    class Meta:
        document = "fragment AssignationEvent on AssignationEvent {\n  id\n  kind\n  returns\n  reference\n  message\n  progress\n}\n\nfragment Assignation on Assignation {\n  args\n  id\n  parent {\n    id\n  }\n  id\n  status\n  events {\n    id\n    returns\n    level\n  }\n  reference\n  updatedAt\n}\n\nfragment AssignationChangeEvent on AssignationChangeEvent {\n  create {\n    ...Assignation\n  }\n  event {\n    ...AssignationEvent\n  }\n}\n\nsubscription WatchAssignations($instanceId: InstanceId!) {\n  assignations(instanceId: $instanceId) {\n    ...AssignationChangeEvent\n  }\n}"


class Get_testcaseQuery(BaseModel):
    test_case: TestCaseFragment = Field(alias="testCase")

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment TestCase on TestCase {\n  id\n  node {\n    id\n  }\n  key\n  isBenchmark\n  description\n  name\n}\n\nquery get_testcase($id: ID!) {\n  testCase(id: $id) {\n    ...TestCase\n  }\n}"


class Get_testresultQuery(BaseModel):
    test_result: TestResultFragment = Field(alias="testResult")

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment TestResult on TestResult {\n  id\n  case {\n    id\n    key\n  }\n  passed\n}\n\nquery get_testresult($id: ID!) {\n  testResult(id: $id) {\n    ...TestResult\n  }\n}"


class Search_testcasesQueryOptions(BaseModel):
    typename: Optional[Literal["TestCase"]] = Field(alias="__typename", exclude=True)
    label: str
    value: ID

    class Config:
        """A config class"""

        frozen = True


class Search_testcasesQuery(BaseModel):
    options: Tuple[Search_testcasesQueryOptions, ...]

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query search_testcases($search: String, $values: [ID!]) {\n  options: testCases(\n    filters: {name: {iContains: $search}, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    label: name\n    value: id\n  }\n}"


class Search_testresultsQueryOptions(BaseModel):
    typename: Optional[Literal["TestResult"]] = Field(alias="__typename", exclude=True)
    label: datetime
    value: ID

    class Config:
        """A config class"""

        frozen = True


class Search_testresultsQuery(BaseModel):
    options: Tuple[Search_testresultsQueryOptions, ...]

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query search_testresults($search: String, $values: [ID!]) {\n  options: testResults(\n    filters: {name: {iContains: $search}, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    label: createdAt\n    value: id\n  }\n}"


class Get_provisionQuery(BaseModel):
    provision: ProvisionFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nfragment Template on Template {\n  id\n  agent {\n    registry {\n      id\n    }\n  }\n  node {\n    ...Node\n  }\n  params\n  extension\n  interface\n}\n\nfragment Provision on Provision {\n  id\n  status\n  template {\n    ...Template\n  }\n}\n\nquery get_provision($id: ID!) {\n  provision(id: $id) {\n    ...Provision\n  }\n}"


class GetMeNodesQueryNodes(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    id: ID
    name: str

    class Config:
        """A config class"""

        frozen = True


class GetMeNodesQuery(BaseModel):
    nodes: Tuple[GetMeNodesQueryNodes, ...]

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "query GetMeNodes {\n  nodes {\n    id\n    name\n  }\n}"


class GetAgentQuery(BaseModel):
    agent: AgentFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment Agent on Agent {\n  registry {\n    app {\n      id\n    }\n    user {\n      id\n    }\n  }\n}\n\nquery GetAgent($id: ID!) {\n  agent(id: $id) {\n    ...Agent\n  }\n}"


class Get_reservationQueryReservationProvisions(BaseModel):
    typename: Optional[Literal["Provision"]] = Field(alias="__typename", exclude=True)
    id: ID
    status: ProvisionEventKind

    class Config:
        """A config class"""

        frozen = True


class Get_reservationQueryReservationNode(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    id: ID
    kind: NodeKind
    name: str

    class Config:
        """A config class"""

        frozen = True


class Get_reservationQueryReservation(BaseModel):
    typename: Optional[Literal["Reservation"]] = Field(alias="__typename", exclude=True)
    id: ID
    provisions: Tuple[Get_reservationQueryReservationProvisions, ...]
    title: Optional[str]
    status: ReservationEventKind
    id: ID
    reference: str
    node: Get_reservationQueryReservationNode

    class Config:
        """A config class"""

        frozen = True


class Get_reservationQuery(BaseModel):
    reservation: Get_reservationQueryReservation

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "query get_reservation($id: ID!) {\n  reservation(id: $id) {\n    id\n    provisions {\n      id\n      status\n    }\n    title\n    status\n    id\n    reference\n    node {\n      id\n      kind\n      name\n    }\n  }\n}"


class ReservationsQuery(BaseModel):
    reservations: Tuple[ReservationFragment, ...]

    class Arguments(BaseModel):
        instance_id: InstanceId

    class Meta:
        document = "fragment Reservation on Reservation {\n  id\n  status\n  node {\n    id\n    hash\n  }\n  waiter {\n    id\n  }\n  reference\n  updatedAt\n}\n\nquery reservations($instance_id: InstanceId!) {\n  reservations(instanceId: $instance_id) {\n    ...Reservation\n  }\n}"


class RequestsQuery(BaseModel):
    assignations: Tuple[AssignationFragment, ...]

    class Arguments(BaseModel):
        instance_id: InstanceId

    class Meta:
        document = "fragment Assignation on Assignation {\n  args\n  id\n  parent {\n    id\n  }\n  id\n  status\n  events {\n    id\n    returns\n    level\n  }\n  reference\n  updatedAt\n}\n\nquery requests($instance_id: InstanceId!) {\n  assignations(instanceId: $instance_id) {\n    ...Assignation\n  }\n}"


class GetEventQuery(BaseModel):
    event: Tuple[AssignationEventFragment, ...]

    class Arguments(BaseModel):
        id: Optional[ID] = Field(default=None)

    class Meta:
        document = "fragment AssignationEvent on AssignationEvent {\n  id\n  kind\n  returns\n  reference\n  message\n  progress\n}\n\nquery GetEvent($id: ID) {\n  event(id: $id) {\n    ...AssignationEvent\n  }\n}"


class Get_templateQuery(BaseModel):
    template: TemplateFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nfragment Template on Template {\n  id\n  agent {\n    registry {\n      id\n    }\n  }\n  node {\n    ...Node\n  }\n  params\n  extension\n  interface\n}\n\nquery get_template($id: ID!) {\n  template(id: $id) {\n    ...Template\n  }\n}"


class Search_templatesQueryOptions(BaseModel):
    typename: Optional[Literal["Template"]] = Field(alias="__typename", exclude=True)
    label: Optional[str]
    value: ID

    class Config:
        """A config class"""

        frozen = True


class Search_templatesQuery(BaseModel):
    options: Tuple[Search_templatesQueryOptions, ...]

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query search_templates($search: String, $values: [ID!]) {\n  options: templates(\n    filters: {interface: {iContains: $search}, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    label: name\n    value: id\n  }\n}"


class FindQuery(BaseModel):
    node: NodeFragment

    class Arguments(BaseModel):
        id: Optional[ID] = Field(default=None)
        template: Optional[ID] = Field(default=None)
        hash: Optional[NodeHash] = Field(default=None)

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nquery find($id: ID, $template: ID, $hash: NodeHash) {\n  node(id: $id, template: $template, hash: $hash) {\n    ...Node\n  }\n}"


class RetrieveallQuery(BaseModel):
    nodes: Tuple[NodeFragment, ...]

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "fragment ChildPortNested on ChildPort {\n  key\n  kind\n  children {\n    identifier\n    nullable\n    kind\n  }\n  identifier\n  nullable\n}\n\nfragment ChildPort on ChildPort {\n  key\n  kind\n  identifier\n  children {\n    ...ChildPortNested\n  }\n  nullable\n}\n\nfragment Port on Port {\n  __typename\n  key\n  label\n  nullable\n  description\n  default\n  kind\n  identifier\n  children {\n    ...ChildPort\n  }\n  validators {\n    function\n    errorMessage\n    dependencies\n    label\n  }\n}\n\nfragment Definition on Node {\n  args {\n    ...Port\n  }\n  returns {\n    ...Port\n  }\n  kind\n  name\n  description\n  collections {\n    name\n  }\n  isDev\n  isTestFor {\n    id\n  }\n  portGroups {\n    key\n  }\n}\n\nfragment Node on Node {\n  hash\n  id\n  ...Definition\n}\n\nquery retrieveall {\n  nodes {\n    ...Node\n  }\n}"


class Search_nodesQueryOptions(Reserve, BaseModel):
    typename: Optional[Literal["Node"]] = Field(alias="__typename", exclude=True)
    label: str
    value: ID

    class Config:
        """A config class"""

        frozen = True


class Search_nodesQuery(BaseModel):
    options: Tuple[Search_nodesQueryOptions, ...]

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query search_nodes($search: String, $values: [ID!]) {\n  options: nodes(\n    filters: {name: {iContains: $search}, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    label: name\n    value: id\n  }\n}"


async def acreate_testcase(
    node: ID,
    key: str,
    description: str,
    name: str,
    is_benchmark: Optional[bool] = None,
    rath: Optional[RekuestNextRath] = None,
) -> TestCaseFragment:
    """create_testcase



    Arguments:
        node (ID): node
        key (str): key
        description (str): description
        name (str): name
        is_benchmark (Optional[bool], optional): is_benchmark.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestCaseFragment"""
    return (
        await aexecute(
            Create_testcaseMutation,
            {
                "node": node,
                "key": key,
                "is_benchmark": is_benchmark,
                "description": description,
                "name": name,
            },
            rath=rath,
        )
    ).create_test_case


def create_testcase(
    node: ID,
    key: str,
    description: str,
    name: str,
    is_benchmark: Optional[bool] = None,
    rath: Optional[RekuestNextRath] = None,
) -> TestCaseFragment:
    """create_testcase



    Arguments:
        node (ID): node
        key (str): key
        description (str): description
        name (str): name
        is_benchmark (Optional[bool], optional): is_benchmark.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestCaseFragment"""
    return execute(
        Create_testcaseMutation,
        {
            "node": node,
            "key": key,
            "is_benchmark": is_benchmark,
            "description": description,
            "name": name,
        },
        rath=rath,
    ).create_test_case


async def acreate_testresult(
    case: ID,
    template: ID,
    passed: bool,
    result: Optional[str] = None,
    rath: Optional[RekuestNextRath] = None,
) -> TestResultFragment:
    """create_testresult



    Arguments:
        case (ID): case
        template (ID): template
        passed (bool): passed
        result (Optional[str], optional): result.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestResultFragment"""
    return (
        await aexecute(
            Create_testresultMutation,
            {"case": case, "template": template, "passed": passed, "result": result},
            rath=rath,
        )
    ).create_test_result


def create_testresult(
    case: ID,
    template: ID,
    passed: bool,
    result: Optional[str] = None,
    rath: Optional[RekuestNextRath] = None,
) -> TestResultFragment:
    """create_testresult



    Arguments:
        case (ID): case
        template (ID): template
        passed (bool): passed
        result (Optional[str], optional): result.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestResultFragment"""
    return execute(
        Create_testresultMutation,
        {"case": case, "template": template, "passed": passed, "result": result},
        rath=rath,
    ).create_test_result


async def aensure_agent(
    instance_id: InstanceId,
    extensions: Optional[List[str]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> EnsureAgentMutationEnsureagent:
    """EnsureAgent



    Arguments:
        instance_id (InstanceId): instanceId
        extensions (Optional[List[str]], optional): extensions.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        EnsureAgentMutationEnsureagent"""
    return (
        await aexecute(
            EnsureAgentMutation,
            {"instanceId": instance_id, "extensions": extensions},
            rath=rath,
        )
    ).ensure_agent


def ensure_agent(
    instance_id: InstanceId,
    extensions: Optional[List[str]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> EnsureAgentMutationEnsureagent:
    """EnsureAgent



    Arguments:
        instance_id (InstanceId): instanceId
        extensions (Optional[List[str]], optional): extensions.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        EnsureAgentMutationEnsureagent"""
    return execute(
        EnsureAgentMutation,
        {"instanceId": instance_id, "extensions": extensions},
        rath=rath,
    ).ensure_agent


async def areserve(
    input: ReserveInput, rath: Optional[RekuestNextRath] = None
) -> ReservationFragment:
    """reserve



    Arguments:
        input (ReserveInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ReservationFragment"""
    return (await aexecute(ReserveMutation, {"input": input}, rath=rath)).reserve


def reserve(
    input: ReserveInput, rath: Optional[RekuestNextRath] = None
) -> ReservationFragment:
    """reserve



    Arguments:
        input (ReserveInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ReservationFragment"""
    return execute(ReserveMutation, {"input": input}, rath=rath).reserve


async def aunreserve(
    input: UnreserveInput, rath: Optional[RekuestNextRath] = None
) -> str:
    """unreserve


     unreserve: The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.


    Arguments:
        input (UnreserveInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        str"""
    return (await aexecute(UnreserveMutation, {"input": input}, rath=rath)).unreserve


def unreserve(input: UnreserveInput, rath: Optional[RekuestNextRath] = None) -> str:
    """unreserve


     unreserve: The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.


    Arguments:
        input (UnreserveInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        str"""
    return execute(UnreserveMutation, {"input": input}, rath=rath).unreserve


async def aassign(
    input: AssignInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """assign



    Arguments:
        input (AssignInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return (await aexecute(AssignMutation, {"input": input}, rath=rath)).assign


def assign(
    input: AssignInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """assign



    Arguments:
        input (AssignInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return execute(AssignMutation, {"input": input}, rath=rath).assign


async def acancel(
    input: CancelInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """cancel



    Arguments:
        input (CancelInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return (await aexecute(CancelMutation, {"input": input}, rath=rath)).cancel


def cancel(
    input: CancelInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """cancel



    Arguments:
        input (CancelInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return execute(CancelMutation, {"input": input}, rath=rath).cancel


async def ainterrupt(
    input: InterruptInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """interrupt



    Arguments:
        input (InterruptInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return (await aexecute(InterruptMutation, {"input": input}, rath=rath)).interrupt


def interrupt(
    input: InterruptInput, rath: Optional[RekuestNextRath] = None
) -> AssignationFragment:
    """interrupt



    Arguments:
        input (InterruptInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationFragment"""
    return execute(InterruptMutation, {"input": input}, rath=rath).interrupt


async def acreate_hardware_record(
    cpu_count: Optional[int] = None,
    cpu_frequency: Optional[float] = None,
    cpu_vendor_name: Optional[str] = None,
    rath: Optional[RekuestNextRath] = None,
) -> CreateHardwareRecordMutationCreatehardwarerecord:
    """CreateHardwareRecord



    Arguments:
        cpu_count (Optional[int], optional): cpuCount.
        cpu_frequency (Optional[float], optional): cpuFrequency.
        cpu_vendor_name (Optional[str], optional): cpuVendorName.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        CreateHardwareRecordMutationCreatehardwarerecord"""
    return (
        await aexecute(
            CreateHardwareRecordMutation,
            {
                "cpuCount": cpu_count,
                "cpuFrequency": cpu_frequency,
                "cpuVendorName": cpu_vendor_name,
            },
            rath=rath,
        )
    ).create_hardware_record


def create_hardware_record(
    cpu_count: Optional[int] = None,
    cpu_frequency: Optional[float] = None,
    cpu_vendor_name: Optional[str] = None,
    rath: Optional[RekuestNextRath] = None,
) -> CreateHardwareRecordMutationCreatehardwarerecord:
    """CreateHardwareRecord



    Arguments:
        cpu_count (Optional[int], optional): cpuCount.
        cpu_frequency (Optional[float], optional): cpuFrequency.
        cpu_vendor_name (Optional[str], optional): cpuVendorName.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        CreateHardwareRecordMutationCreatehardwarerecord"""
    return execute(
        CreateHardwareRecordMutation,
        {
            "cpuCount": cpu_count,
            "cpuFrequency": cpu_frequency,
            "cpuVendorName": cpu_vendor_name,
        },
        rath=rath,
    ).create_hardware_record


async def acreate_template(
    input: CreateTemplateInput, rath: Optional[RekuestNextRath] = None
) -> TemplateFragment:
    """createTemplate



    Arguments:
        input (CreateTemplateInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TemplateFragment"""
    return (
        await aexecute(CreateTemplateMutation, {"input": input}, rath=rath)
    ).create_template


def create_template(
    input: CreateTemplateInput, rath: Optional[RekuestNextRath] = None
) -> TemplateFragment:
    """createTemplate



    Arguments:
        input (CreateTemplateInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TemplateFragment"""
    return execute(CreateTemplateMutation, {"input": input}, rath=rath).create_template


async def aset_extension_templates(
    input: SetExtensionTemplatesInput, rath: Optional[RekuestNextRath] = None
) -> List[TemplateFragment]:
    """SetExtensionTemplates



    Arguments:
        input (SetExtensionTemplatesInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[TemplateFragment]"""
    return (
        await aexecute(SetExtensionTemplatesMutation, {"input": input}, rath=rath)
    ).set_extension_templates


def set_extension_templates(
    input: SetExtensionTemplatesInput, rath: Optional[RekuestNextRath] = None
) -> List[TemplateFragment]:
    """SetExtensionTemplates



    Arguments:
        input (SetExtensionTemplatesInput): input
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[TemplateFragment]"""
    return execute(
        SetExtensionTemplatesMutation, {"input": input}, rath=rath
    ).set_extension_templates


async def awatch_reservations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> AsyncIterator[ReservationFragment]:
    """WatchReservations



    Arguments:
        instance_id (InstanceId): instanceId
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ReservationFragment"""
    async for event in asubscribe(
        WatchReservationsSubscription, {"instanceId": instance_id}, rath=rath
    ):
        yield event.reservations


def watch_reservations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> Iterator[ReservationFragment]:
    """WatchReservations



    Arguments:
        instance_id (InstanceId): instanceId
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ReservationFragment"""
    for event in subscribe(
        WatchReservationsSubscription, {"instanceId": instance_id}, rath=rath
    ):
        yield event.reservations


async def awatch_assignations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> AsyncIterator[AssignationChangeEventFragment]:
    """WatchAssignations



    Arguments:
        instance_id (InstanceId): instanceId
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationChangeEventFragment"""
    async for event in asubscribe(
        WatchAssignationsSubscription, {"instanceId": instance_id}, rath=rath
    ):
        yield event.assignations


def watch_assignations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> Iterator[AssignationChangeEventFragment]:
    """WatchAssignations



    Arguments:
        instance_id (InstanceId): instanceId
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AssignationChangeEventFragment"""
    for event in subscribe(
        WatchAssignationsSubscription, {"instanceId": instance_id}, rath=rath
    ):
        yield event.assignations


async def aget_testcase(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> TestCaseFragment:
    """get_testcase



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestCaseFragment"""
    return (await aexecute(Get_testcaseQuery, {"id": id}, rath=rath)).test_case


def get_testcase(id: ID, rath: Optional[RekuestNextRath] = None) -> TestCaseFragment:
    """get_testcase



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestCaseFragment"""
    return execute(Get_testcaseQuery, {"id": id}, rath=rath).test_case


async def aget_testresult(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> TestResultFragment:
    """get_testresult



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestResultFragment"""
    return (await aexecute(Get_testresultQuery, {"id": id}, rath=rath)).test_result


def get_testresult(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> TestResultFragment:
    """get_testresult



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TestResultFragment"""
    return execute(Get_testresultQuery, {"id": id}, rath=rath).test_result


async def asearch_testcases(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_testcasesQueryOptions]:
    """search_testcases



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_testcasesQueryTestcases]"""
    return (
        await aexecute(
            Search_testcasesQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_testcases(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_testcasesQueryOptions]:
    """search_testcases



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_testcasesQueryTestcases]"""
    return execute(
        Search_testcasesQuery, {"search": search, "values": values}, rath=rath
    ).options


async def asearch_testresults(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_testresultsQueryOptions]:
    """search_testresults



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_testresultsQueryTestresults]"""
    return (
        await aexecute(
            Search_testresultsQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_testresults(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_testresultsQueryOptions]:
    """search_testresults



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_testresultsQueryTestresults]"""
    return execute(
        Search_testresultsQuery, {"search": search, "values": values}, rath=rath
    ).options


async def aget_provision(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> ProvisionFragment:
    """get_provision



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ProvisionFragment"""
    return (await aexecute(Get_provisionQuery, {"id": id}, rath=rath)).provision


def get_provision(id: ID, rath: Optional[RekuestNextRath] = None) -> ProvisionFragment:
    """get_provision



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        ProvisionFragment"""
    return execute(Get_provisionQuery, {"id": id}, rath=rath).provision


async def aget_me_nodes(
    rath: Optional[RekuestNextRath] = None,
) -> List[GetMeNodesQueryNodes]:
    """GetMeNodes



    Arguments:
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[GetMeNodesQueryNodes]"""
    return (await aexecute(GetMeNodesQuery, {}, rath=rath)).nodes


def get_me_nodes(rath: Optional[RekuestNextRath] = None) -> List[GetMeNodesQueryNodes]:
    """GetMeNodes



    Arguments:
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[GetMeNodesQueryNodes]"""
    return execute(GetMeNodesQuery, {}, rath=rath).nodes


async def aget_agent(id: ID, rath: Optional[RekuestNextRath] = None) -> AgentFragment:
    """GetAgent



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AgentFragment"""
    return (await aexecute(GetAgentQuery, {"id": id}, rath=rath)).agent


def get_agent(id: ID, rath: Optional[RekuestNextRath] = None) -> AgentFragment:
    """GetAgent



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        AgentFragment"""
    return execute(GetAgentQuery, {"id": id}, rath=rath).agent


async def aget_reservation(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> Get_reservationQueryReservation:
    """get_reservation



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        Get_reservationQueryReservation"""
    return (await aexecute(Get_reservationQuery, {"id": id}, rath=rath)).reservation


def get_reservation(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> Get_reservationQueryReservation:
    """get_reservation



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        Get_reservationQueryReservation"""
    return execute(Get_reservationQuery, {"id": id}, rath=rath).reservation


async def areservations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> List[ReservationFragment]:
    """reservations



    Arguments:
        instance_id (InstanceId): instance_id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[ReservationFragment]"""
    return (
        await aexecute(ReservationsQuery, {"instance_id": instance_id}, rath=rath)
    ).reservations


def reservations(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> List[ReservationFragment]:
    """reservations



    Arguments:
        instance_id (InstanceId): instance_id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[ReservationFragment]"""
    return execute(
        ReservationsQuery, {"instance_id": instance_id}, rath=rath
    ).reservations


async def arequests(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> List[AssignationFragment]:
    """requests



    Arguments:
        instance_id (InstanceId): instance_id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[AssignationFragment]"""
    return (
        await aexecute(RequestsQuery, {"instance_id": instance_id}, rath=rath)
    ).assignations


def requests(
    instance_id: InstanceId, rath: Optional[RekuestNextRath] = None
) -> List[AssignationFragment]:
    """requests



    Arguments:
        instance_id (InstanceId): instance_id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[AssignationFragment]"""
    return execute(RequestsQuery, {"instance_id": instance_id}, rath=rath).assignations


async def aget_event(
    id: Optional[ID] = None, rath: Optional[RekuestNextRath] = None
) -> List[AssignationEventFragment]:
    """GetEvent



    Arguments:
        id (Optional[ID], optional): id.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[AssignationEventFragment]"""
    return (await aexecute(GetEventQuery, {"id": id}, rath=rath)).event


def get_event(
    id: Optional[ID] = None, rath: Optional[RekuestNextRath] = None
) -> List[AssignationEventFragment]:
    """GetEvent



    Arguments:
        id (Optional[ID], optional): id.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[AssignationEventFragment]"""
    return execute(GetEventQuery, {"id": id}, rath=rath).event


async def aget_template(
    id: ID, rath: Optional[RekuestNextRath] = None
) -> TemplateFragment:
    """get_template



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TemplateFragment"""
    return (await aexecute(Get_templateQuery, {"id": id}, rath=rath)).template


def get_template(id: ID, rath: Optional[RekuestNextRath] = None) -> TemplateFragment:
    """get_template



    Arguments:
        id (ID): id
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        TemplateFragment"""
    return execute(Get_templateQuery, {"id": id}, rath=rath).template


async def asearch_templates(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_templatesQueryOptions]:
    """search_templates



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_templatesQueryTemplates]"""
    return (
        await aexecute(
            Search_templatesQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_templates(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_templatesQueryOptions]:
    """search_templates



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_templatesQueryTemplates]"""
    return execute(
        Search_templatesQuery, {"search": search, "values": values}, rath=rath
    ).options


async def afind(
    id: Optional[ID] = None,
    template: Optional[ID] = None,
    hash: Optional[NodeHash] = None,
    rath: Optional[RekuestNextRath] = None,
) -> NodeFragment:
    """find



    Arguments:
        id (Optional[ID], optional): id.
        template (Optional[ID], optional): template.
        hash (Optional[NodeHash], optional): hash.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        NodeFragment"""
    return (
        await aexecute(
            FindQuery, {"id": id, "template": template, "hash": hash}, rath=rath
        )
    ).node


def find(
    id: Optional[ID] = None,
    template: Optional[ID] = None,
    hash: Optional[NodeHash] = None,
    rath: Optional[RekuestNextRath] = None,
) -> NodeFragment:
    """find



    Arguments:
        id (Optional[ID], optional): id.
        template (Optional[ID], optional): template.
        hash (Optional[NodeHash], optional): hash.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        NodeFragment"""
    return execute(
        FindQuery, {"id": id, "template": template, "hash": hash}, rath=rath
    ).node


async def aretrieveall(rath: Optional[RekuestNextRath] = None) -> List[NodeFragment]:
    """retrieveall



    Arguments:
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[NodeFragment]"""
    return (await aexecute(RetrieveallQuery, {}, rath=rath)).nodes


def retrieveall(rath: Optional[RekuestNextRath] = None) -> List[NodeFragment]:
    """retrieveall



    Arguments:
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[NodeFragment]"""
    return execute(RetrieveallQuery, {}, rath=rath).nodes


async def asearch_nodes(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_nodesQueryOptions]:
    """search_nodes



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_nodesQueryNodes]"""
    return (
        await aexecute(
            Search_nodesQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_nodes(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[RekuestNextRath] = None,
) -> List[Search_nodesQueryOptions]:
    """search_nodes



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (rekuest_next.rath.RekuestNextRath, optional): The arkitekt rath client

    Returns:
        List[Search_nodesQueryNodes]"""
    return execute(
        Search_nodesQuery, {"search": search, "values": values}, rath=rath
    ).options


AssignInput.update_forward_refs()
AssignWidgetInput.update_forward_refs()
ChildPortInput.update_forward_refs()
CreateTemplateInput.update_forward_refs()
DefinitionInput.update_forward_refs()
DependencyInput.update_forward_refs()
EffectInput.update_forward_refs()
PortInput.update_forward_refs()
ProvisionFragment.update_forward_refs()
TemplateFragment.update_forward_refs()
TemplateInput.update_forward_refs()



import asyncio
import logging
import uuid
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from pydantic import BaseModel, Field, PrivateAttr

from koil.types import Contextual
from rekuest_next.actors.errors import ProvisionDelegateException, UnknownMessageError
from rekuest_next.actors.transport.local_transport import ProxyActorTransport
from rekuest_next.actors.transport.types import ActorTransport, AssignTransport
from rekuest_next.actors.types import Assignment, Passport, Unassignment
from rekuest_next.api.schema import (
    AssignationEventKind,
    LogLevel,
    ProvisionEventKind,
    TemplateFragment,
)
from rekuest_next.collection.collector import (
    ActorCollector,
    AssignationCollector,
    Collector,
)
from rekuest_next.definition.define import DefinitionInput
from rekuest_next.definition.registry import (
    DefinitionRegistry,
)
from rekuest_next.messages import OutMessage
from rekuest_next.structures.registry import (
    StructureRegistry,
)

logger = logging.getLogger(__name__)


@runtime_checkable
class Agent(Protocol):
    async def abuild_actor_for_template(
        self, template: TemplateFragment, passport: Passport, transport: ActorTransport
    ) -> "Actor": ...


class Actor(BaseModel):
    passport: Passport
    transport: ActorTransport
    collector: Collector
    agent: Agent
    supervisor: Optional["Actor"] = None
    managed_actors: Dict[str, "Actor"] = Field(default_factory=dict)
    running_assignments: Dict[str, Assignment] = Field(default_factory=dict)

    _in_queue: Optional[asyncio.Queue] = PrivateAttr(default=None)
    _running_asyncio_tasks: Dict[str, asyncio.Task] = PrivateAttr(default_factory=dict)
    _running_transports: Dict[str, AssignTransport] = PrivateAttr(default_factory=dict)
    _provision_task: asyncio.Task = PrivateAttr(default=None)
    _status: ProvisionEventKind = PrivateAttr(default=ProvisionEventKind.UNHAPPY)

    async def on_provide(self, passport: Passport):
        return None

    async def on_unprovide(self):
        return None

    async def on_assign(
        self,
        assignment: Assignment,
        collector: AssignationCollector,
        transport: AssignTransport,
    ):
        raise NotImplementedError(
            "Needs to be owerwritten in Actor Subclass. Never use this class directly"
        )

    async def aget_status(self):
        return self._status

    async def apass(self, message: Union[Unassignment, Assignment]):
        assert self._in_queue, "Actor is currently not listening"
        await self._in_queue.put(message)

    async def arun(self):
        self._in_queue = asyncio.Queue()
        self._provision_task = asyncio.create_task(self.alisten())
        return self._provision_task

    async def acancel(self):
        # Cancel Mnaged actors
        logger.info(f"Cancelling Actor {self.passport.id}")
        if self.managed_actors:
            logger.info("Cancelling Actors")
            for actor in self.managed_actors.values():
                logger.info(
                    f"Cancelling managed actor with passport {actor.passport.id}"
                )
                await actor.acancel()

        if not self._provision_task or self._provision_task.done():
            # Race condition
            return

        self._provision_task.cancel()

        try:
            await self._provision_task
        except asyncio.CancelledError:
            logger.info(f"Actor {self.passport.id} was cancelled")

    async def provide(self):
        try:
            logging.info(f"Providing {self.passport}")
            await self.on_provide(self.passport)
            await self.transport.log_event(
                kind=ProvisionEventKind.ACTIVE,
                message=f"Wonderfully provided {self.passport}",
            )

        except Exception as e:
            logging.critical(f"Providing Error {self.passport} {e}", exc_info=True)
            await self.transport.log_event(
                kind=ProvisionEventKind.DENIED,
                message=f"Providing {self.passport}",
            )

    async def unprovide(self):
        try:
            await self.on_unprovide()
            await self.transport.log_event(
                kind=ProvisionEventKind.UNHAPPY,
                message=f"Providing {self.passport}",
            )

        except Exception as e:
            logging.critical(f"Unproviding Error {self.passport} {e}", exc_info=True)
            await self.transport.log_event(
                kind=ProvisionEventKind.CRITICAL,
                message=f"Unproviding Error {self.passport} {e}",
            )

    def assign_task_done(self, task):
        logger.info(f"Assign task is done: {task}")
        if task.exception():
            try:
                raise task.exception()
            except asyncio.CancelledError:
                pass
            except Exception as e:
                logger.error(
                    f"Assign task {task} failed with exception {e}", exc_info=True
                )
            pass

    async def is_assignment_still_running(self, message: Union[Assignment]):
        if message.id in self._running_asyncio_tasks:
            return True
        return False

    async def aprocess(self, message: Union[Assignment, Unassignment]):
        logger.info(f"Actor for {self.passport}: Received {message}")

        if isinstance(message, Assignment):
            transport = self.transport.spawn(message)

            task = asyncio.create_task(
                self.on_assign(
                    message,
                    collector=self.collector,
                    transport=transport,
                )
            )

            task.add_done_callback(self.assign_task_done)

            self._running_transports[message.id] = transport
            self._running_asyncio_tasks[message.id] = task

        elif isinstance(message, Unassignment):
            if message.id in self._running_asyncio_tasks:
                task = self._running_asyncio_tasks[message.id]
                assign_transport = self._running_transports[message.id]

                if not task.done():
                    task.cancel()
                    try:
                        await task
                    except asyncio.CancelledError:
                        logger.info(
                            f"Task {assign_transport.assignment} was cancelled through arkitekt. Setting Cancelled"
                        )
                        await assign_transport.log_event(
                            kind=AssignationEventKind.CANCELLED,
                            message="Cancelled remotely",
                        )
                        del self._running_asyncio_tasks[message.id]
                        del self._running_transports[message.id]

                else:
                    logger.warning(
                        f"Race Condition: Task was already done before cancellation"
                    )
                    await assign_transport.log_event(
                        kind=AssignationEventKind.ERROR,
                        message=f"Race Condition: Task was already done before cancellation",
                    )

            else:
                logger.error(
                    f"Actor for {self.passport}: Received unassignment for unknown assignation {message.id}"
                )
        else:
            raise UnknownMessageError(f"{message}")

    async def alisten(self):
        try:
            await self.provide()
            logger.info(f"Actor for {self.passport}: Is now active")

            while True:
                message = await self._in_queue.get()
                try:
                    await self.aprocess(message)
                except Exception as e:
                    logger.critical(
                        "Processing unknown message should never happen", exc_info=True
                    )

        except asyncio.CancelledError:
            logger.info("Doing Whatever needs to be done to cancel!")

            [i.cancel() for i in self._running_asyncio_tasks.values()]

            for task, assign_transport in zip(
                self._running_asyncio_tasks.values(), self._running_transports.values()
            ):
                try:
                    await task
                except asyncio.CancelledError:
                    logger.info(
                        f"Task {assign_transport.assignment} was cancelled through applicaction. Setting Critical"
                    )
                    await assign_transport.log_event(
                        kind=AssignationEventKind.CRITICAL,
                        message="Cancelled trhough application (this is not nice from the application and will be regarded as an error)",
                    )

            await self.unprovide()

        except Exception as e:
            logger.critical("Unhandled exception", exc_info=True)

            # TODO: Maybe send back an acknoledgement that we are done cancelling.
            # If we don't do this, arkitekt will not know if we failed to cancel our
            # tasks or if we succeeded. If we fail to cancel arkitekt can try to
            # kill the whole agent (maybe causing a sys.exit(1) or something)

        self._in_queue = None

    def _provision_task_done(self, task):
        logger.info(f"Provision task is done: {task}")
        if task.exception():
            raise task.exception()

    async def __aenter__(self):
        return self

    async def aspawn_actor(
        self,
        template: TemplateFragment,
        on_log_event: Callable[[OutMessage], Awaitable[None]],
    ) -> "SerializingActor":
        """Spawns an Actor managed by thisfrom the definition of the given interface"""

        passport = Passport(
            provision=self.passport.provision,
            parent=self.passport.id,
            instance_id=self.passport.instance_id,
        )

        transport = ProxyActorTransport(passport=passport, on_log_event=on_log_event)

        actor = await self.agent.abuild_actor_for_template(
            template,
            passport,
            transport,
        )

        self.managed_actors[actor.passport.id] = actor
        return actor

    async def __aexit__(self, exc_type, exc, tb):
        if self._provision_task and not self._provision_task.done():
            await self.acancel()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        copy_on_model_validation = "none"


class SerializingActor(Actor):
    definition: DefinitionInput
    structure_registry: StructureRegistry
    expand_inputs: bool = True
    shrink_outputs: bool = True


Actor.update_forward_refs()
SerializingActor.update_forward_refs()



from pydantic import BaseModel, Field
from typing import Dict, Any
import contextvars
import uuid

current_shelve = contextvars.ContextVar("current_shelve", default=None)
GLOBAL_SHELVE = None


def get_default_definition_registry():
    global GLOBAL_SHELVE
    if GLOBAL_SHELVE is None:
        GLOBAL_SHELVE = Shelve()
    return GLOBAL_SHELVE


def get_current_shelve(allow_global=True):
    return current_shelve.get(get_default_definition_registry())


class Shelve(BaseModel):
    store: Dict[str, Any] = Field(default_factory=dict)
    free_on_exit: bool = False

    async def aput(self, data: Any):
        """
        Put data in the store.

        :param key: The key to store the data under.
        :param data: The data to store.
        :return: None
        """
        key = str(uuid.uuid4())
        self.store[key] = data
        return key

    async def aget(self, key: str):
        return self.store[key]

    async def adelete(self, key: str):
        del self.store[key]

    async def __aenter__(self):
        self._token = current_shelve.set(self)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.free_on_exit:
            # Freeing the store on exit is useful for testing.
            self.store.clear()

        current_shelve.set(
            None
        )  # should be self._token but that gives an incorrect loop error
        return False



import contextvars
from rekuest_next.actors.helper import AssignmentHelper
from rekuest_next.actors.errors import (
    NotWithinAnAssignationError,
)

current_assignment = contextvars.ContextVar("current_assignment")
current_assignation_helper = contextvars.ContextVar("assignment_helper")


def get_current_assignation_helper() -> AssignmentHelper:
    try:
        return current_assignation_helper.get()
    except LookupError as e:
        raise NotWithinAnAssignationError(
            "Trying to access assignation helper outside of an assignation"
        ) from e



from rekuest_next.errors import RekuestError


class ActorException(RekuestError):
    """An exception that is raised by an actor"""

    pass


class UnknownMessageError(ActorException):
    """An exception that is raised when an actor receives an unknown message"""

    pass


class ThreadedActorCancelled(ActorException):
    """An exception that is raised when a threaded actor is cancelled"""

    pass


class ContextError(ActorException):
    """An exception that is raised when an cannot access the context
    (probably because it is not within one)"""

    pass


class NotWithinAnAssignationError(ContextError):
    """Indicates that an actors task is not within an assignation"""

    pass


class NotWithinAProvisionError(ContextError):
    """Indicates that an actors task is not within a provision"""

    pass


class ProvisionDelegateException(ActorException):
    """An exception that is raised when trying to delageted actor"""


class DelegateException(ActorException):
    """An exception that is raised by a delageted actor"""



from rekuest_next.api.schema import (
    DependencyInput,
    LogicalCondition,
    EffectInput,
    EffectKind,
)
from typing import Any


def hiddenwhile(key: str, condition: LogicalCondition, value: Any) -> EffectInput:
    return EffectInput(
        dependencies=[DependencyInput(key=key, condition=condition, value=value)],
        kind=EffectKind.HIDDEN,
    )


def hiddenif(key: str) -> EffectInput:
    return EffectInput(
        dependencies=[
            DependencyInput(key=key, condition=LogicalCondition.IS, value=True)
        ],
        kind=EffectKind.HIDDEN,
    )


def showif(key: str) -> EffectInput:
    return EffectInput(
        dependencies=[
            DependencyInput(key=key, condition=LogicalCondition.IS_NOT, value=True)
        ],
        kind=EffectKind.HIDDEN,
    )


def neverdothis(condition: LogicalCondition, value: Any) -> EffectInput:
    """No one should ever do this

    Why are you even reading this?

    Args:
        condition (LogicalCondition): Tell me why you are reading this
        value (Any): This is not a joke

    Returns:
        EffectInput: A stupid effect
    """
    return EffectInput(
        dependencies=[DependencyInput(condition=condition, value=value)],
        kind=EffectKind.CRAZY,
    )



from rekuest_next.rath import RekuestNextRath, current_rekuest_next_rath
from koil import unkoil
from koil.helpers import unkoil_gen


def execute(operation, variables, rath: RekuestNextRath = None):
    return unkoil(aexecute, operation, variables, rath)


async def aexecute(operation, variables, rath: RekuestNextRath = None):
    rath = rath or current_rekuest_next_rath.get()
    x = await rath.aquery(
        operation.Meta.document, operation.Arguments(**variables).dict(by_alias=True)
    )
    return operation(**x.data)


def subscribe(operation, variables, rath: RekuestNextRath = None):
    """is subscribing"""
    return unkoil_gen(asubscribe, operation, variables, rath)


async def asubscribe(operation, variables, rath: RekuestNextRath = None):
    rath = rath or current_rekuest_next_rath.get()
    async for event in rath.asubscribe(
        operation.Meta.document, operation.Arguments(**variables).dict(by_alias=True)
    ):
        yield operation(**event.data)



from typing import Dict
from pydantic import Field
from rekuest_next.api.schema import TemplateFragment
from rekuest_next.postmans.graphql import GraphQLPostman
from rekuest_next.rath import RekuestNextRath
from rekuest_next.structures.default import get_default_structure_registry
from rekuest_next.structures.registry import (
    StructureRegistry,
)

from rekuest_next.definition.registry import (
    DefinitionRegistry,
    get_current_definition_registry,
    get_default_definition_registry,
)
from rekuest_next.agents.base import BaseAgent
from rekuest_next.postmans.base import BasePostman
from koil import unkoil
from koil.composition import Composition
from koil.decorators import koilable
from rekuest_next.register import register


@koilable(fieldname="koil", add_connectors=True)
class RekuestNext(Composition):
    rath: RekuestNextRath = Field(default_factory=RekuestNextRath)
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )
    agent: BaseAgent = Field(default_factory=BaseAgent)
    postman: BasePostman = Field(default_factory=GraphQLPostman)

    registered_templates: Dict[str, TemplateFragment] = Field(default_factory=dict)

    def register(self, *args, **kwargs) -> None:
        """
        Register a new function
        """
        structure_registry = kwargs.pop("structure_registry", self.structure_registry)
        definition_registry = kwargs.pop(
            "definition_registry", self.agent.definition_registry
        )

        return register(
            *args,
            definition_registry=definition_registry,
            structure_registry=structure_registry,
            **kwargs,
        )

    def run(self, *args, **kwargs) -> None:
        """
        Run the application.
        """
        return unkoil(self.arun, *args, **kwargs)

    async def arun(self) -> None:
        """
        Run the application.
        """
        await self.agent.aprovide()

    def _repr_html_inline_(self):
        return f"<table><tr><td>rath</td><td>{self.rath._repr_html_inline_()}</td></tr></table>"

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        extra = "forbid"



from typing import Protocol, runtime_checkable, Callable, Awaitable, Any
from rekuest_next.structures.registry import StructureRegistry
from rekuest_next.api.schema import PortGroupInput, AssignationEventKind
from rekuest_next.definition.define import DefinitionInput
from typing import Optional, List, Dict, Tuple, Any
from pydantic import BaseModel, Field
import uuid



class Passport(BaseModel):
    instance_id: str
    provision: str
    parent: Optional[str]
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))


class Assignment(BaseModel):
    assignation: Optional[str]
    parent: Optional[str]
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    args: Dict[str, Any] = Field(default_factory=dict)
    user: Optional[str]
    reference: Optional[str]
    context: Dict[str, Any]  = Field(default_factory=dict)


class AssignmentUpdate(BaseModel):
    assignment: str
    status: AssignationEventKind
    message: Optional[str]
    parent: Optional[str]
    progress: Optional[int]
    returns: Optional[List[Any]]


class Unassignment(BaseModel):
    assignation: str
    id: str
    context: Dict[str, Any]  = Field(default_factory=dict)

@runtime_checkable
class ActorBuilder(Protocol):
    def __call__(
        self,
        passport: Passport,
        transport: Any,
        collector: Any,
        definition_registry: Any,
    ) -> Any: ...


@runtime_checkable
class Actifier(Protocol):
    """An actifier is a function that takes a callable and a structure registry
    as well as optional arguments

    """

    def __call__(
        self,
        function: Callable,
        structure_registry: StructureRegistry,
        port_groups: Optional[List[PortGroupInput]] = None,
        groups: Optional[Dict[str, List[str]]] = None,
        is_test_for: Optional[List[str]] = None,
        **kwargs,
    ) -> Tuple[DefinitionInput, ActorBuilder]: ...


@runtime_checkable
class OnProvide(Protocol):
    """An on_provide is a function that takes a provision and a transport and returns
    an awaitable

    """

    def __call__(
        self,
        passport: Passport,
    ) -> Awaitable[Any]: ...


@runtime_checkable
class OnUnprovide(Protocol):
    """An on_provide is a function that takes a provision and a transport and returns
    an awaitable

    """

    def __call__(self) -> Awaitable[Any]: ...



from dataclasses import dataclass
from rekuest_next.messages import Assign
from rekuest_next.actors.helper import AssignmentHelper
from rekuest_next.actors.vars import (
    current_assignation_helper,
    current_assignment,
)
from rekuest_next.actors.transport.types import AssignTransport
from rekuest_next.actors.types import Assignment, Passport
from pydantic import BaseModel


class AssignmentContext(BaseModel):
    passport: Passport
    assignment: Assignment
    transport: AssignTransport
    _helper = None

    def __enter__(self):
        self._helper = AssignmentHelper(
            assignment=self.assignment, transport=self.transport, passport=self.passport
        )

        current_assignment.set(self.assignment)
        current_assignation_helper.set(self._helper)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        current_assignation_helper.set(None)
        current_assignment.set(None)
        self._helper = None

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__exit__(exc_type, exc_val, exc_tb)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from rekuest_next.actors.types import Passport, Assignment
from rekuest_next.structures.default import (
    get_default_structure_registry,
    StructureRegistry,
)
from pydantic import BaseModel, Field
from typing import Dict, Any, List
import logging


logger = logging.getLogger(__name__)


class AssignationCollector(BaseModel):
    assignment: Assignment

    async def collect(self):
        """
        Collect data from the source.

        :return: The collected data.
        """
        raise NotImplementedError

    async def register(self, args: List[Any]):
        return


class ActorCollector(BaseModel):
    """
    The ActorCollector class is used to collect data in the course of a
    an anctors life time. It initiates Assignation that then in turn
    collect data during the Assignation liftetime
    """

    passport: Passport
    sub_collectors: Dict[str, AssignationCollector] = Field(default_factory=dict)
    delegated_collector: Dict[str, "ActorCollector"] = Field(default_factory=dict)

    def spawn(self, assignment: Assignment):
        """
        Spawn a new collector for the given assignation.
        """
        assign_collector = AssignationCollector(assignment=assignment)
        self.sub_collectors[assignment.parent] = assign_collector
        return assign_collector

    async def collect(self, assignment: Assignment):
        """
        Collect data from the source.

        :return: The collected data.
        """
        raise NotImplementedError


class Collector(BaseModel):
    """
    The Collector class is used to collect data in the course of a
    an agents life time. It initiates ActorCollectors that then in turn
    collect data during the actors liftetime
    """

    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )

    assignment_map: Dict[str, List[Any]] = Field(default_factory=dict)
    children_tree: Dict[str, List[str]] = Field(default_factory=dict)

    def register(self, assignment: Assignment, items: List[any]):
        logger.debug(f"Registering {assignment.id}")

        if assignment.id in self.assignment_map:
            self.assignment_map[assignment.id] += items
        else:
            self.assignment_map[assignment.id] = items
        if assignment.parent:
            if assignment.parent in self.children_tree:
                self.children_tree[assignment.parent].append(assignment.id)
            else:
                self.children_tree[assignment.parent] = [assignment.id]

    async def collect(self, id: str):
        """
        Collect data from the source.

        :return: The collected data.
        """

        if id in self.assignment_map:
            for identifier, value in self.assignment_map[id]:
                try:
                    collector = self.structure_registry.get_collector_for_identifier(
                        identifier
                    )
                    await collector(value)
                except Exception as e:
                    logger.critical(
                        f"Error while collecting {identifier} with value {value}"
                    )
                    raise e

        if id in self.children_tree:
            for child in self.children_tree[id]:
                await self.collect(child)

    class Config:
        copy_on_model_validation = False



from typing import AsyncIterator, Awaitable, Callable, Optional

from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.links.errors import AuthenticationError
from rath.errors import NotComposedError
from rekuest_next.actors.vars import get_current_assignation_helper


class ContextLink(ContinuationLink):
    """ContextLink is a link that adds an assignation token to the context.
    The authentication token is retrieved by calling the token_loader function.
    If the wrapped link raises an AuthenticationError, the token_refresher function
    is called again to refresh the token.

    This link is statelss, and does not store the token. It is up to the user to
    store the token and pass it to the token_loader function.
    """

    async def aexecute(
        self, operation: Operation, retry: int = 0
    ) -> AsyncIterator[GraphQLResult]:
        """Executes and forwards an operation to the next link.

        This method will add the authentication token to the context of the operation,
        and will refresh the token if the next link raises an AuthenticationError, until
        the maximum number of refresh attempts is reached.

        Parameters
        ----------
        operation : Operation
            The operation to execute

        Yields
        ------
        GraphQLResult
            The result of the operation
        """
        if not self.next:
            raise NotComposedError("No next link set")

        try:
            helper = get_current_assignation_helper()

            operation.context.headers["x-assignation-id"] = helper.assignation
            print("Assignation", helper.assignation)
        except Exception as e:
            print("Error", e)

        async for result in self.next.aexecute(operation):
            yield result

    class Config:
        """pydantic configuration for the AuthTokenLink"""

        underscore_attrs_are_private = True
        arbitary_types_allowed = True



from rekuest_next.api.schema import DefinitionInput, PortKind
from typing import List, Any


def parse_collectable(definition: DefinitionInput, returns: List[Any]):
    collectables = []

    if returns is None:
        returns = []
    elif not isinstance(returns, tuple):
        returns = [returns]

    for port, v in zip(definition.returns, returns):
        # Only structures will be collected because others are passed by value
        if port.kind == PortKind.STRUCTURE:
            collectables.append((port.identifier, v))

    return collectables



from typing import Any, Dict, List, Optional, Tuple, Union
from rekuest_next.api.schema import NodeFragment
import asyncio
from rekuest_next.structures.errors import ExpandingError, ShrinkingError
from rekuest_next.structures.registry import StructureRegistry
from rekuest_next.api.schema import (
    PortFragment,
    PortKind,
    DefinitionInput,
    DefinitionFragment,
    ChildPortFragment,
)
from rekuest_next.structures.errors import (
    PortShrinkingError,
    StructureShrinkingError,
    PortExpandingError,
    StructureExpandingError,
)
from .predication import predicate_port
import datetime as dt


async def ashrink_arg(
    port: Union[PortFragment, ChildPortFragment], value: Any, structure_registry=None
) -> Any:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    try:
        if value is None:
            if port.nullable:
                return None
            else:
                raise ValueError(
                    "{port} is not nullable (optional) but your provided None"
                )

        if port.kind == PortKind.DICT:
            return {
                key: await ashrink_arg(port.children[0], value, structure_registry)
                for key, value in value.items()
            }

        if port.kind == PortKind.LIST:
            return await asyncio.gather(
                *[
                    ashrink_arg(port.children[0], item, structure_registry=structure_registry)
                    for item in value
                ]
            )

        if port.kind == PortKind.INT:
            return int(value) if value is not None else None

        if port.kind == PortKind.UNION:
            for index, x in enumerate(port.children):
                if predicate_port(x, value, structure_registry):
                    return {
                        "use": index,
                        "value": await ashrink_arg(x, value, structure_registry),
                    }

            raise ShrinkingError(
                f"Port is union butn none of the predicated for this port held true {port.children}"
            )

        if port.kind == PortKind.DATE:
            return value.isoformat() if value is not None else None

        if port.kind == PortKind.STRUCTURE:
            # We always convert structures returns to strings
            try:
                shrinker = structure_registry.get_shrinker_for_identifier(
                    port.identifier
                )
            except KeyError:
                raise StructureShrinkingError(
                    f"Couldn't find shrinker for {port.identifier}"
                ) from None
            try:
                shrink = await shrinker(value)
                return str(shrink)
            except Exception:
                raise StructureShrinkingError(
                    f"Error shrinking {repr(value)} with Structure {port.identifier}"
                ) from None

        if port.kind == PortKind.BOOL:
            return bool(value) if value is not None else None

        if port.kind == PortKind.STRING:
            return str(value) if value is not None else None

        raise NotImplementedError(f"Should be implemented by subclass {port}")

    except Exception as e:
        raise PortShrinkingError(
            f"Couldn't shrink value {value} with port {port}"
        ) from e


async def ashrink_args(
    node: NodeFragment,
    args: List[Any],
    kwargs: Dict[str, Any],
    structure_registry: StructureRegistry,
) -> Dict[str, Any]:
    """Shrinks args and kwargs

    Shrinks the inputs according to the Node Definition

    Args:
        node (Node): The Node

    Raises:
        ShrinkingError: If args are not Shrinkable
        ShrinkingError: If kwargs are not Shrinkable

    Returns:
        Tuple[List[Any], Dict[str, Any]]: Parsed Args as a List, Parsed Kwargs as a dict
    """

    try:
        args_iterator = iter(args)
    except TypeError:
        raise ShrinkingError(f"Couldn't iterate over args {args}")

    # Extract to Argslist

    shrinked_kwargs = {}

    for port in node.args:
        try:
            arg = next(args_iterator)
        except StopIteration as e:
            if port.key in kwargs:
                arg = kwargs[port.key]
            else:
                if port.nullable or port.default is not None:
                    arg = None  # defaults will be set by the agent
                else:
                    raise ShrinkingError(
                        f"Couldn't find value for nonnunllable port {port.key}"
                    ) from e

        try:
            shrunk_arg = await ashrink_arg(port, arg, structure_registry)
            shrinked_kwargs[port.key] = shrunk_arg
        except Exception as e:
            raise ShrinkingError(f"Couldn't shrink arg {arg} with port {port}") from e

    return shrinked_kwargs


async def aexpand_return(
    port: Union[PortFragment, ChildPortFragment],
    value: Any,
    structure_registry=None,
) -> Any:
    """Expand a value through a port

    Args:
        port (ArgPortFragment): Port to expand to
        value (Any): Value to expand
    Returns:
        Any: Expanded value

    """
    if value is None:
        if port.nullable:
            return None
        else:
            raise PortExpandingError(
                f"{port} is not nullable (optional) but your provided None"
            )

    if port.kind == PortKind.DICT:
        return {
            key: await aexpand_return(port.child, value, structure_registry)
            for key, value in value.items()
        }

    if port.kind == PortKind.LIST:
        return await asyncio.gather(
            *[
                aexpand_return(port.child, item, structure_registry=structure_registry)
                for item in value
            ]
        )

    if port.kind == PortKind.UNION:
        assert isinstance(value, dict), "Union value needs to be a dict"
        assert "use" in value, "No use in vaalue"
        index = value["use"]
        true_value = value["value"]
        return await aexpand_return(
            port.variants[index], true_value, structure_registry=structure_registry
        )

    if port.kind == PortKind.INT:
        return int(value)

    if port.kind == PortKind.FLOAT:
        return float(value)

    if port.kind == PortKind.DATE:
        return dt.datetime.fromisoformat(value.replace("Z", "+00:00"))

    if port.kind == PortKind.STRUCTURE:
        if not (isinstance(value, str) or isinstance(value, int)):
            raise PortExpandingError(
                f"Expected value to be a string or int, but got {type(value)}"
            )

        try:
            expander = structure_registry.get_expander_for_identifier(port.identifier)
        except KeyError:
            raise StructureExpandingError(
                f"Couldn't find expander for {port.identifier}"
            ) from None

        try:
            return await expander(value)
        except Exception:
            raise StructureExpandingError(
                f"Error expanding {repr(value)} with Structure {port.identifier}"
            ) from None

    if port.kind == PortKind.BOOL:
        return bool(value)

    if port.kind == PortKind.STRING:
        return str(value)

    raise NotImplementedError("Should be implemented by subclass")


async def aexpand_returns(
    node: NodeFragment,
    returns: Dict[str, Any],
    structure_registry: StructureRegistry,
) -> Tuple[Any]:
    """Expands Returns

    Expands the Returns according to the Node definition


    Args:
        node (Node): Node definition
        returns (List[any]): The returns

    Raises:
        ExpandingError: if they are not expandable

    Returns:
        List[Any]: The Expanded Returns
    """
    assert returns is not None, "Returns can't be empty"

    expanded_returns = []

    for port in node.returns:
        expanded_return = None
        if port.key not in returns:
            if port.nullable:
                returns[port.key] = None
            else:
                raise ExpandingError(f"Missing key {port.key} in returns")

        else:
            try:
                expanded_return = await aexpand_return(
                    port, returns[port.key], structure_registry
                )
            except Exception as e:
                raise ExpandingError(
                    f"Couldn't expand return {returns[port.key]} with port {port}"
                ) from e

        expanded_returns.append(expanded_return)

    return expanded_return


def serialize_inputs(
    definition: Union[DefinitionFragment, DefinitionInput],
    kwargs: Dict[str, Any],
) -> Tuple[Any]:
    """Shrinks args and kwargs

    Shrinks the inputs according to the Node Definition

    Args:
        node (Node): The Node

    Raises:
        ShrinkingError: If args are not Shrinkable
        ShrinkingError: If kwargs are not Shrinkable

    Returns:
        Tuple[List[Any], Dict[str, Any]]: Parsed Args as a List, Parsed Kwargs as a dict
    """

    args_list = []

    # Extract to Argslist

    for port in definition.args:
        value = kwargs.pop(port.key, None)
        if value is None and not port.nullable:
            raise ShrinkingError(
                f"Couldn't find value for nonnunllable port {port.key}"
            )
        args_list.append(value)

    shrinked_args = args_list

    return tuple(shrinked_args)


def deserialize_outputs(
    definition: Union[DefinitionFragment, DefinitionInput],
    returns: List[Any],
) -> Dict[str, Any]:
    """Expands Returns

    Expands the Returns according to the Node definition


    Args:
        node (Node): Node definition
        returns (List[any]): The returns

    Raises:
        ExpandingError: if they are not expandable

    Returns:
        Dcit[str, Any]: The Expanded Returns
    """
    assert returns is not None, "Returns can't be empty"
    if len(definition.returns) != len(returns):
        raise ExpandingError(
            f"Missmatch in Return Length. Node requires {len(definition.returns)} returns,"
            f" but got {len(returns)}"
        )

    values = {}

    for port, value in zip(definition.returns, returns):
        values[port.key] = value

    return values



import asyncio
import logging
from typing import Any, Dict, List, Optional, Protocol, Union, runtime_checkable

from pydantic import BaseModel, Field, PrivateAttr

from koil.types import Contextual
from rekuest_next.actors.types import Assignment, Passport
from rekuest_next.api.schema import ProvisionEventKind, AssignationEventKind, LogLevel


@runtime_checkable
class ActorTransport(Protocol):
    passport: Passport

    async def log_event(
        self,
        kind: ProvisionEventKind = None,
        message: str = None,
        level: LogLevel = None,
    ): ...

    def spawn(self, assignment: Assignment) -> "AssignTransport": ...


@runtime_checkable
class AssignTransport(Protocol):
    assignment: Assignment

    async def log_event(
        self,
        kind: AssignationEventKind = None,
        message: str = None,
        returns: List[Any] = None,
        progress: int = None,
    ): ...
